# #!/usr/bin/python3
# -*- coding: utf-8 -*-


import copy


import numpy as np
from bsplines2d import BSplines2D as Previous


# from . import _class00_check as _check
from . import _class00_check_ion as _check_ion
from . import _class00_check_lines as _check_lines
from . import _class00_compute as _compute
from . import _class00_lines_labels as _labels
from . import _class00_plot as _plot
from . import _class00_plot_spectral_data as _plot_spectral_data


__all__ = ['SpectralLines']


#############################################
#############################################
#       DEFAULT VALUES
#############################################


_QUANT_NE = 'ne'
_QUANT_TE = 'Te'
_UNITS_LAMBDA0 = 'm'


#############################################
#############################################
#       Spectral Lines
#############################################


class SpectralLines(Previous):

    _ddef = copy.deepcopy(Previous._ddef)

    # _show_in_summary_core = ['shape', 'ref', 'group']
    _dshow = dict(Previous._dshow)
    _dshow['ion'] = ['element', 'A', 'Z', 'q', 'isoelect']

    _ddef['params']['dobj'] = {
        # 'ion': {
            # 'A': {'cls': float},
            # 'Z': {'cls': int},
            # 'q': {'cls': int},
            # 'isoelect': {'cls': str},
        # },
        'lines': {
            'lambda0': {'cls': float, 'def': 0.},
            'source': {'cls': str, 'def': 'unknown'},
            'transition': {'cls': str, 'def': 'unknown'},
            'element':  {'cls': str, 'def': 'unknown'},
            'charge':  {'cls': int, 'def': 0},
            'ion':  {'cls': str, 'def': 'unknown'},
            'symbol':   {'cls': str, 'def': 'unknown'},
        },
    }

    _which_ion = 'ion'
    _which_lines = 'spect_lines'
    _quant_ne = _QUANT_NE
    _quant_Te = _QUANT_TE

    _units_lamb0 = _UNITS_LAMBDA0

    # -------------------
    # add ion
    # -------------------

    def add_ion(self, key=None):
        """ Add an ion
        """
        _check_ion.add_ion(
            coll=self,
            key=key,
        )

    def remove_ion(self, key=None, propagate=None):
        """ Add an ion
        """
        _check_ion.remove_ion(
            coll=self,
            key=key,
            propagate=propagate,
        )

    # -------------------
    # add spectral lines
    # -------------------

    def add_spectral_line(
        self,
        key=None,
        ion=None,
        lamb0=None,
        transition=None,
        pec=None,
        source=None,
        symbol=None,
        **kwdargs,
    ):
        """ Add a spectral line by key and rest wavelength, optionally with

        """
        _check_lines.add_lines(
            coll=self,
            key=key,
            lamb0=lamb0,
            pec=pec,
            source=source,
            transition=transition,
            ion=ion,
            symbol=symbol,
            **kwdargs,
        )

    def add_spectral_lines_from_file(
        self,
        pfe=None,
    ):
        """ Add a spectral line from json, npz or py file

        """
        dobj = _compute.from_file(
            coll=self,
            pfe=pfe,
        )

        return _check_lines._add_lines_from_dobj(
            coll=self,
            dobj=dobj,
        )

    # ---------------------
    # Add PEC
    # ---------------------

    def add_pec(self, key=None, pec=None, ref=None):
        pass

    # -----------------
    # from openadas
    # ------------------

    def add_spectral_lines_from_openadas(
        self,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        online=None,
        update=None,
        create_custom=None,
    ):
        """
        Load lines and pec from openadas, either:
            - online = True:  directly from the website
            - online = False: from pre-downloaded files in ~/.spectrally/openadas/

        Provide wavelengths in m

        Example:
        --------
                >>> import spectrally as sp
                >>> coll = sp.Collection()
                >>> coll.add_spectral_lines_from_openadas(
                    element='Mo',
                    lambmin=3.94e-10,
                    lambmax=4e-10,
                )

        """

        ddata, dref, dobj = _compute.from_openadas(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            online=online,
            update=update,
            create_custom=create_custom,
            dsource0=self._dobj.get('source'),
            dref0=self._dref if len(self._dref) > 0 else None,
            ddata0=self._ddata if len(self._ddata) > 0 else None,
            dobj0=self._dobj,
            which_lines=self._which_lines,
        )

        return _check_lines._add_lines_from_dobj(
            coll=self,
            dref=dref,
            ddata=ddata,
            dobj=dobj,
        )

    # -----------------
    # from nist
    # ------------------

    def add_spectral_lines_from_nist(
        self,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        ion=None,
        wav_observed=None,
        wav_calculated=None,
        transitions_allowed=None,
        transitions_forbidden=None,
        cache_from=None,
        cache_info=None,
        verb=None,
        create_custom=None,
    ):
        """
        Load lines from nist, either:
            - cache_from = False:  directly from the website
            - cache_from = True: from pre-downloaded files in ~/.spectrally/nist/

        Provide wavelengths in m

        Example:
        --------
                >>> import spectrally as sp
                >>> coll = sp.Collection()
                >>> sp.add_spectral_lines_from_nist(
                    element='Mo',
                    lambmin=3.94e-10,
                    lambmax=4e-10,
                )

        """

        dobj = _compute.from_nist(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            ion=ion,
            wav_observed=wav_observed,
            wav_calculated=wav_calculated,
            transitions_allowed=transitions_allowed,
            transitions_forbidden=transitions_forbidden,
            cache_from=cache_from,
            cache_info=cache_info,
            verb=verb,
            create_custom=create_custom,
            dsource0=self._dobj.get('source'),
            dlines0=self._dobj.get('lines'),
            group_lines=self._which_lines,
        )

        return _check_lines._add_lines_from_dobj(
            coll=self,
            dobj=dobj,
        )

    # ---------------
    # remove lines
    # ---------------

    def remove_spectral_lines(self, keys=None, propagate=None):
        return _check_lines.remove_lines(
            coll=self,
            keys=keys,
            propagate=propagate,
        )

    # -----------------
    # conversion wavelength - energy - frequency
    # ------------------

    def convert_lines(self, units=None, key=None, ind=None, returnas=None):
        """ Convert wavelength (m) to other units or other quantities

        Avalaible units:
            wavelength: km, m, mm, um, nm, pm, A
            energy:     J, eV, keV, MeV, GeV
            frequency:  Hz, kHz, MHz, GHz, THz

        Return the result as a np.ndarray (returnas = 'data')

        Can also just return the conversion coef if returnas='coef'
        In that case, a bool is also returned indicating whether the result is
        the proportional to the inverse of lambda0::
            - False: data = coef * lambda0
            - True: data = coef / lambda0
        """
        if units is None:
            units = self._units_lambda0
        if returnas is None:
            returnas = dict

        lok = [dict, np.ndarray, 'data', 'coef']
        if returnas not in lok:
            msg = (
                "Arg returnas must be in:\n"
                + "\t- {}\n".format(lok)
                + "\t- provided: {}".format(returnas)
            )
            raise Exception(msg)
        if returnas in [dict, np.ndarray, 'data']:
            returnas2 = 'data'
        else:
            returnas2 = 'coef'

        # get keys of desired lines
        key = self._ind_tofrom_key(
            which=self._which_lines, key=key, ind=ind, returnas=str,
        )

        # get wavelength in m
        lamb_in = self.get_param(
            which=self._which_lines, param='lambda0',
            key=key, returnas=np.ndarray,
        )['lambda0']

        # conversion
        out = _compute.convert_spectral(
            data_in=lamb_in,
            units_in='m',
            units_out=units,
            returnas=returnas2,
        )
        if returnas is dict:
            out = {k0: out[ii] for ii, k0 in enumerate(key)}
        return out

    # ---------------
    # get labels
    # ---------------

    def get_spectral_lines_labels(
        self,
        keys=None,
        labels=None,
        colors=None,
    ):
        """ Return a dict of {key: label}

        Typically used to customize spectral lines labelling

        If keys is a key to a spectral model or spectral fit, the spectral
        lines are actually the names of all model functions that can be peaked,
        which includes the following:
                - gauss
                - lorentz
                - pvoigt
                - pulse_exp
                - pulse_gauss

        Parameters
        ----------
        keys:     str / list of str
            Can be either:
                - key of list of keys of valid spectral lines
                - key of a valid spectral model
                - key of a valid spectral fit

        labels:   str / dict
            Can be either:
                - str: a valid spectral line parameter
                    e.g.: 'symbol', 'ion', ...
                - dict: a custom dict of {key: label}

        Returns
        -------
        dlabels:  dict
            dict of {key: label}

        """
        return _labels.main(
            coll=self,
            keys=keys,
            labels=labels,
            colors=colors,
        )

    # ---------------
    # plot spctral data
    # ---------------

    def plot_spectral_data(
        self,
        key_data=None,
        key_lamb=None,
        keyY=None,
        key_lines=None,
        # plotting
        dprop=None,
        dvminmax=None,
        # lines labels
        lines_labels=True,
        lines_labels_color=None,
        lines_labels_rotation=None,
        lines_labels_horizontalalignment=None,
        # figure
        dax=None,
        fs=None,
        dmargin=None,
        tit=None,
        # interactivity
        nmax=None,
        connect=None,
        dinc=None,
        show_commands=None,
    ):
        """ Plot a 1d or 2d array overlaid with spectral lines labelling

        """

        return _plot_spectral_data.main(
            coll=self,
            key_data=key_data,
            key_lamb=key_lamb,
            keyY=keyY,
            key_lines=key_lines,
            # plotting
            dprop=dprop,
            dvminmax=dvminmax,
            # lines labels
            lines_labels=lines_labels,
            lines_labels_color=lines_labels_color,
            lines_labels_rotation=lines_labels_rotation,
            lines_labels_horizontalalignment=lines_labels_horizontalalignment,
            # figure
            dax=dax,
            fs=fs,
            dmargin=dmargin,
            tit=tit,
            # interactivity
            nmax=nmax,
            connect=connect,
            dinc=dinc,
            show_commands=show_commands,
        )

    # -----------------
    # PEC interpolation
    # ------------------

# =============================================================================
#     def calc_pec(
#         self,
#         key=None,
#         ind=None,
#         ne=None,
#         Te=None,
#         deg=None,
#         grid=None,
#         return_params=None,
#     ):
#         """ Compute the pec (<sigma v>) by interpolation for chosen lines
#
#         Assumes Maxwellian electron distribution
#
#         Provide ne and Te and 1d np.ndarrays
#
#         if grid=False:
#             - ne is a (n,) 1d array
#             - Te is a (n,) 1d array
#           => the result is a dict of (n,) 1d array
#
#         if grid=True:
#             - ne is a (n1,) 1d array
#             - Te is a (n2,) 1d array
#           => the result is a dict of (n1, n2) 2d array
#         """
#
#         # Check keys
#         key = self._ind_tofrom_key(
#             which=self._which_lines, key=key, ind=ind, returnas=str,
#         )
#         dlines = self._dobj[self._which_lines]
#
#         key, dnTe, return_params = _check._check_compute_pec(
#             # check keys
#             key=key,
#             dlines=dlines,
#             ddata=self._ddata,
#             _quant_ne=self._quant_ne,
#             _quant_Te=self._quant_Te,
#             # check ne, Te
#             ne=ne,
#             Te=Te,
#             return_params=return_params,
#         )
#         keypec = [f'{k0}-pec' for k0 in key]
#
#         # group lines per common ref
#         lref = set([self._ddata[k0]['ref'] for k0 in keypec])
#         dref = {
#             k0: [k1 for k1 in keypec if self._ddata[k1]['ref'] == k0]
#             for k0 in lref
#         }
#
#         # Interpolate
#         for ii, (k0, v0) in enumerate(dref.items()):
#             douti, dparami = self.interpolate(
#                 # interpolation base
#                 keys=v0,
#                 ref_key=k0,
#                 # interpolation pts
#                 x0=dnTe['ne'],
#                 x1=dnTe['Te'],
#                 # parameters
#                 deg=deg,
#                 deriv=0,
#                 grid=grid,
#                 log_log=True,
#                 return_params=True,
#             )
#
#             # update dict
#             if ii == 0:
#                 dout = douti
#                 dparam = dparami
#             else:
#                 dout.update(**douti)
#                 dparam['keys'] += dparami['keys']
#                 dparam['ref_key'] += dparami['ref_key']
#
#         # -------
#         # return
#
#         if return_params is True:
#             dparam['key'] = dparam['keys']
#             del dparam['keys']
#             dparam['ne'] = dparam['x0']
#             dparam['Te'] = dparam['x1']
#             del dparam['x0'], dparam['x1']
#             return dout, dparam
#         else:
#             return dout
#
#     def calc_intensity(
#         self,
#         key=None,
#         ind=None,
#         ne=None,
#         Te=None,
#         concentration=None,
#         deg=None,
#         grid=None,
#     ):
#         """ Compute the lines intensities by pec interpolation for chosen lines
#
#         Assumes Maxwellian electron distribution
#         Assumes concentration = nz / ne
#
#         Provide ne and Te and 1d np.ndarrays
#
#         Provide concentration as:
#             - a np.ndarray (same concentration assumed for all lines)
#             - a dict of {key: np.ndarray}
#
#         if grid=False:
#             - ne is a (n,) 1d array
#             - Te is a (n,) 1d array
#             - concentration is a (dict of) (n,) 1d array(s)
#           => the result is a dict of (n1, n2) 2d array
#
#         if grid=True:
#             - ne is a (n1,) 1d array
#             - Te is a (n2,) 1d array
#             - concentration is a (dict of) (n1, n2) 2d array(s)
#           => the result is a dict of (n1, n2) 2d array
#
#
#         """
#
#         # Check keys
#         key = self._ind_tofrom_key(
#             which=self._which_lines, key=key, ind=ind, returnas=str,
#         )
#
#         # interpolate pec
#         dout, dparam = self.calc_pec(
#             key=key,
#             ind=ind,
#             ne=ne,
#             Te=Te,
#             grid=grid,
#             deg=deg,
#             return_params=True,
#         )
#
#         # check concentrations
#         concentration = _check._check_compute_intensity(
#             key=[k0[:-4] for k0 in dparam['key']],
#             concentration=concentration,
#             shape=dparam['ne'].shape,
#         )
#
#         # Derive intensity
#         for k0, v0 in dout.items():
#             dout[k0] = v0['data']*dparam['ne']**2*concentration[k0[:-4]]
#
#         return dout
# =============================================================================

    # -----------------
    # plotting
    # ------------------

    def plot_spectral_lines(
        self,
        key=None,
        ind=None,
        ax=None,
        sortby=None,
        param_x=None,
        param_txt=None,
        ymin=None,
        ymax=None,
        ls=None,
        lw=None,
        fontsize=None,
        side=None,
        dsize=None,
        dcolor=None,
        fraction=None,
        figsize=None,
        dmargin=None,
        wintit=None,
        tit=None,
    ):
        """ plot rest wavelengths as vertical lines """

        # Check inputs
        key = self._ind_tofrom_key(
            which=self._which_lines, key=key, ind=ind, returnas=str,
        )

        return _plot.plot_axvlines(
            din=self._dobj[self._which_lines],
            key=key,
            param_x=param_x,
            param_txt=param_txt,
            sortby=sortby,
            dsize=dsize,
            ax=ax, ymin=ymin, ymax=ymax,
            ls=ls, lw=lw, fontsize=fontsize,
            side=side, dcolor=dcolor, fraction=fraction,
            figsize=figsize, dmargin=dmargin,
            wintit=wintit, tit=tit,
        )

# =============================================================================
#     def plot_pec_single(
#         self,
#         key=None,
#         ind=None,
#         ne=None,
#         Te=None,
#         concentration=None,
#         deg=None,
#         grid=None,
#         ax=None,
#         sortby=None,
#         param_x=None,
#         param_txt=None,
#         ymin=None,
#         ymax=None,
#         ls=None,
#         lw=None,
#         fontsize=None,
#         side=None,
#         dcolor=None,
#         fraction=None,
#         figsize=None,
#         dmargin=None,
#         wintit=None,
#         tit=None,
#     ):
#         """ Same as plot_spectral_lines() with extra scatter plot with circles
#
#         The circles' diameters depend on the pec value for each line
#
#         Requires:
#             - Te = scalar (eV)
#             - ne = scalar (/m3)
#
#         """
#
#         # ------------
#         # Check ne, Te
#
#         ltypes = [int, float, np.integer, np.floating]
#         dnTe = {'ne': ne, 'Te': Te}
#         single = all([
#             type(v0) in ltypes or len(v0) == 1 for v0 in dnTe.values()
#         ])
#         if not single:
#             msg = ("Arg ne and Te must be floats!")
#             raise Exception(msg)
#
#         # --------
#         # Get dpec
#         dpec = self.calc_pec(
#             key=key,
#             ind=ind,
#             ne=ne,
#             Te=Te,
#             deg=deg,
#             grid=False,
#             return_params=False,
#         )
#         key = [k0[:-4] for k0 in dpec.keys()]
#
#         ne = float(ne)
#         Te = float(Te)
#         tit = (
#             r'$n_e$' + f'= {ne} ' + r'$/m^3$'
#             + r' -  $T_e$ = ' + f'{Te/1000.} keV'
#         )
#
#         pmax = np.max([np.log10(v0['data']) for v0 in dpec.values()])
#         pmin = np.min([np.log10(v0['data']) for v0 in dpec.values()])
#         dsize = {
#             k0[:-4]: (np.log10(v0['data']) - pmin) / (pmax - pmin)*19 + 1
#             for k0, v0 in dpec.items()
#         }
#
#         return _plot.plot_axvlines(
#             din=self._dobj[self._which_lines],
#             key=key,
#             param_x=param_x,
#             param_txt=param_txt,
#             sortby=sortby,
#             dsize=dsize,
#             ax=ax, ymin=ymin, ymax=ymax,
#             ls=ls, lw=lw, fontsize=fontsize,
#             side=side, dcolor=dcolor, fraction=fraction,
#             figsize=figsize, dmargin=dmargin,
#             wintit=wintit, tit=tit,
#         )
#
#     def plot_pec(
#         self,
#         key=None,
#         ind=None,
#         ne=None,
#         Te=None,
#         norder=None,
#         ne_scale=None,
#         Te_scale=None,
#         param_txt=None,
#         param_color=None,
#         deg=None,
#         dax=None,
#         proj=None,
#         ymin=None,
#         ymax=None,
#         ls=None,
#         lw=None,
#         fontsize=None,
#         side=None,
#         dcolor=None,
#         fraction=None,
#         figsize=None,
#         dmargin=None,
#         dtit=None,
#         tit=None,
#         wintit=None,
#     ):
#
#         # -----------
#         # Check input
#
#         # Check ne, Te
#         lc = [np.isscalar(ne) or len(ne) == 1, np.isscalar(Te) or len(Te) == 1]
#         if all(lc):
#             msg = "For a single (ne, Te) space, use plot_pec_singe()"
#             raise Exception(msg)
#
#         # Get dpec
#         dpec = self.calc_pec(
#             key=key,
#             ind=ind,
#             ne=ne,
#             Te=Te,
#             deg=deg,
#             grid=False,
#         )
#         damp = {k0: {'data': v0} for k0, v0 in dpec.items()}
#
#         # Create grid
#         ne_grid = ds._class1_compute._get_grid1d(
#             ne, scale=ne_scale, npts=ne.size*2, nptsmin=3,
#         )
#         Te_grid = ds._class1_compute._get_grid1d(
#             Te, scale=Te_scale, npts=Te.size*2, nptsmin=3,
#         )
#
#         # get dpec for grid
#         dpec_grid = self.calc_pec(
#             key=key,
#             ind=ind,
#             ne=ne_grid,
#             Te=Te_grid,
#             deg=deg,
#             grid=True,
#         )
#
#         raise NotImplementedError()
#
#         return _plot.plot_dominance_map(
#             din=self._dobj['lines'], im=im, extent=extent,
#             xval=ne, yval=Te, damp=damp,
#             x_scale=ne_scale, y_scale=Te_scale, amp_scale='log',
#             param_txt='symbol',
#             dcolor=dcolor,
#             dax=dax, proj=proj,
#             figsize=figsize, dmargin=dmargin,
#             wintit=wintit, tit=tit, dtit=dtit,
#         )
# =============================================================================

    # -------------------
    # units conversion
    # -------------------

# =============================================================================
#     def convert_spectral_lines_units(
#         self,
#         data=None,
#         units=None,
#         units_in=None,
#     ):
#
#         return _compute.convert_spectral_units(
#             coll=self,
#             data=data,
#             units=units,
#             units_in=units_in,
#         )
# =============================================================================

    # -------------------
    # saving
    # -------------------

    def save_spectral_lines_to_file(
        self,
        path=None,
        name=None,
        keys=None,
        overwrite=None,
    ):
        """ Save a set of spectral lines as a json file (or npz) """

        return _check_lines._save_to_file(
            coll=self,
            keys=keys,
            name=name,
            path=path,
            overwrite=overwrite,
        )


# #!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
created on tue feb 20 14:44:51 2024

@author: dvezinet
"""


import numpy as np
import scipy.special as scpsp


# ############################################
# ############################################
#       details
# ############################################


def _get_func_details(
    c0=None,
    c1=None,
    c2=None,
    dind=None,
    param_val=None,
):

    # --------------
    # prepare
    # --------------

    def func(
        x_free=None,
        lamb=None,
        param_val=param_val,
        c0=c0,
        c1=c1,
        c2=c2,
        dind=dind,
        scales=None,
        iok=None,
        bin_iok=None,
        # binning
        bin_ind=None,
        bin_dlamb=None,
    ):

        # ---------------------
        # get lamb limits + iok

        # for pulses
        lamb00 = lamb[0]
        lambD = lamb[-1] - lamb[0]
        lambm = 0.5*(lamb[-1] + lamb[0])

        # iok
        if bin_iok is not None and bin_ind is False:
            lamb = lamb[bin_iok]

        # ----------
        # initialize

        shape = tuple([dind['nfunc']] + list(lamb.shape))
        val = np.zeros(shape, dtype=float)

        # -------------------
        # rescale

        if scales is not None:
            x_free = x_free * scales

        # ----------------------------
        # get x_full from constraints

        if c0 is None:
            x_full = x_free
        else:
            x_full = c2.dot(x_free**2) + c1.dot(x_free) + c0

        # ------------------
        # sum all poly

        kfunc = 'poly'
        if dind.get(kfunc) is not None:

            a0 = x_full[dind[kfunc]['a0']['ind']][:, None]
            a1 = x_full[dind[kfunc]['a1']['ind']][:, None]
            a2 = x_full[dind[kfunc]['a2']['ind']][:, None]

            ind = dind['func'][kfunc]['ind']
            lamb_rel = (lamb - lambm) / lambD
            val[ind, ...] = (
                a0
                + a1 * lamb_rel
                + a2 * lamb_rel**2
            )

        # --------------------
        # sum all exponentials

        kfunc = 'exp_lamb'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][:, None]
            rate = x_full[dind[kfunc]['rate']['ind']][:, None]

            ind = dind['func'][kfunc]['ind']
            val[ind, ...] = amp * np.exp(- rate / lamb) / lamb

        # -----------------
        # sum all gaussians

        kfunc = 'gauss'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][:, None]
            sigma = x_full[dind[kfunc]['sigma']['ind']][:, None]
            vccos = x_full[dind[kfunc]['vccos']['ind']][:, None]
            lamb0 = param_val[dind[kfunc]['lamb0']][:, None]

            ind = dind['func'][kfunc]['ind']
            val[ind, ...] = (
                amp * np.exp(-(lamb - lamb0*(1 + vccos))**2/(2*sigma**2))
            )

        # -------------------
        # sum all Lorentzians

        kfunc = 'lorentz'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][:, None]
            gam = x_full[dind[kfunc]['gam']['ind']][:, None]
            vccos = x_full[dind[kfunc]['vccos']['ind']][:, None]
            lamb0 = param_val[dind[kfunc]['lamb0']][:, None]

            # https://en.wikipedia.org/wiki/Cauchy_distribution
            # value at lamb0 = amp / (pi * gam)

            ind = dind['func'][kfunc]['ind']
            val[ind, ...] = (
                amp / (1 + ((lamb - lamb0*(1 + vccos)) / gam)**2)
            )

        # --------------------
        # sum all pseudo-voigt

        kfunc = 'pvoigt'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][:, None]
            sigma = x_full[dind[kfunc]['sigma']['ind']][:, None]
            gam = x_full[dind[kfunc]['gam']['ind']][:, None]
            vccos = x_full[dind[kfunc]['vccos']['ind']][:, None]
            lamb0 = param_val[dind[kfunc]['lamb0']][:, None]

            # https://en.wikipedia.org/wiki/Voigt_profile

            fg = 2 * np.sqrt(2*np.log(2)) * sigma
            fl = 2 * gam
            ftot = (
                fg**5 + 2.69269*fg**4*fl + 2.42843*fg**3*fl**2
                + 4.47163*fg**2*fl**3 + 0.07842*fg*fl**4 + fl**5
            ) ** (1./5.)
            ratio = fl / ftot

            # eta
            eta = 1.36603 * ratio - 0.47719 * ratio**2 + 0.11116 * ratio**3

            # update widths of gauss and Lorentz
            sigma2 = ftot / (2 * np.sqrt(2*np.log(2)))
            gam2 = ftot / 2.

            # weighted sum
            ind = dind['func'][kfunc]['ind']
            val[ind, ...] = amp * (
                eta / (1 + ((lamb - lamb0*(1 + vccos)) / gam2)**2)
                + (1-eta) * np.exp(
                    -(lamb - lamb0*(1 + vccos))**2
                    / (2*sigma2**2)
                )
            )

        # ------------

        kfunc = 'voigt'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][:, None]
            sigma = x_full[dind[kfunc]['sigma']['ind']][:, None]
            gam = x_full[dind[kfunc]['gam']['ind']][:, None]
            vccos = x_full[dind[kfunc]['vccos']['ind']][:, None]
            lamb0 = param_val[dind[kfunc]['lamb0']][:, None]

            ind = dind['func'][kfunc]['ind']
            val[ind, ...] = amp * scpsp.voigt_profile(
                lamb - lamb0*(1 + vccos),
                sigma,
                gam,
            )

        # ------------------
        # sum all pulse_exp

        kfunc = 'pulse_exp'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][:, None]
            tau = x_full[dind[kfunc]['tau']['ind']][:, None]
            tup = x_full[dind[kfunc]['t_up']['ind']][:, None]
            tdown = x_full[dind[kfunc]['t_down']['ind']][:, None]

            ind0 = lamb > (lamb00 + lambD * tau)
            dlamb = lamb - (lamb00 + lambD * tau)

            ind = dind['func'][kfunc]['ind']
            val[ind, ...] = (
                amp * ind0 * (
                    np.exp(-dlamb/tdown)
                    - np.exp(-dlamb/tup)
                )
            )

        # ------------------
        # sum all pulse_gauss

        kfunc = 'pulse_gauss'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][:, None]
            tau = x_full[dind[kfunc]['tau']['ind']][:, None]
            tup = x_full[dind[kfunc]['t_up']['ind']][:, None]
            tdown = x_full[dind[kfunc]['t_down']['ind']][:, None]

            indup = (lamb < (lamb00 + lambD * tau))
            inddown = (lamb >= (lamb00 + lambD * tau))

            ind = dind['func'][kfunc]['ind']
            dlamb = lamb - (lamb00 + lambD * tau)

            val[ind, ...] = (
                amp * (
                    indup * np.exp(-dlamb**2/tup**2)
                    + inddown * np.exp(-dlamb**2/tdown**2)
                )
            )

        # ------------------
        # sum all lognorm

        kfunc = 'lognorm'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']]
            tau = x_full[dind[kfunc]['tau']['ind']]
            sigma = x_full[dind[kfunc]['sigma']['ind']]
            mu = x_full[dind[kfunc]['mu']['ind']]

            # max at t - t0 = exp(mu - sigma**2)
            # max = amp * exp(sigma**2/2 - mu)
            # variance = (exp(sigma**2) - 1) * exp(2mu + sigma**2)
            # skewness = (exp(sigma**2) + 2) * sqrt(exp(sigma**2) - 1)

            ind = dind['func'][kfunc]['ind']
            for ii, i0 in enumerate(ind):
                ioki = lamb > (lamb00 + lambD * tau[ii])

                dlamb = lamb[ioki] - (lamb00 + lambD * tau[ii])

                val[i0, ioki] = (
                    (amp[ii] / dlamb)
                    * np.exp(-(np.log(dlamb) - mu[ii])**2 / (2.*sigma[ii]**2))
                )

        # -------
        # binning

        if bin_ind is not False:

            val = np.add.reduceat(val * bin_dlamb[None, :], bin_ind, axis=1)
            if iok is not None:
                val = val[:, iok]

        return val

    return func


# ############################################
# ############################################
#       sum
# ############################################


def _get_func_sum(
    c0=None,
    c1=None,
    c2=None,
    dind=None,
    param_val=None,
):

    # --------------
    # prepare
    # --------------

    func_details = _get_func_details(
        c0=c0,
        c1=c1,
        c2=c2,
        dind=dind,
        param_val=param_val,
    )

    # --------------
    # prepare
    # --------------

    def func(
        x_free=None,
        lamb=None,
        # scales, iok
        scales=None,
        iok=None,
        bin_iok=None,
        bin_ind=None,
        bin_dlamb=None,
        # unused (data)
        **kwdargs,
    ):

        return np.sum(
            func_details(
                x_free,
                lamb=lamb,
                scales=scales,
                iok=iok,
                bin_iok=bin_iok,
                bin_ind=bin_ind,
                bin_dlamb=bin_dlamb,
            ),
            axis=0,
        )

    return func


# ############################################
# ############################################
#       cost
# ############################################


def _get_func_cost(
    c0=None,
    c1=None,
    c2=None,
    dind=None,
    param_val=None,
):

    # --------------
    # prepare
    # --------------

    func_sum = _get_func_sum(
        c0=c0,
        c1=c1,
        c2=c2,
        dind=dind,
        param_val=param_val,
    )

    # ------------
    # cost
    # ------------

    def func(
        x_free=None,
        lamb=None,
        # scales, iok
        scales=None,
        iok=None,
        # binning
        bin_ind=None,
        bin_iok=None,
        bin_dlamb=None,
        # data
        data=None,
        # sum
        func_sum=func_sum,
    ):
        if iok is not None:
            data = data[iok]

        return func_sum(
            x_free,
            lamb=lamb,
            scales=scales,
            iok=iok,
            bin_iok=bin_iok,
            bin_ind=bin_ind,
            bin_dlamb=bin_dlamb,
        ) - data

    return func


# ############################################
# ############################################
#       Jacobian
# ############################################


def _get_func_jacob(
    c0=None,
    c1=None,
    c2=None,
    dindj=None,
    dind=None,
    param_val=None,
):

    # --------------
    # prepare
    # --------------

    def func(
        x_free=None,
        lamb=None,
        param_val=param_val,
        c0=c0,
        c1=c1,
        c2=c2,
        dindj=dindj,
        dind=dind,
        # scales, iok
        scales=None,
        iok=None,
        bin_iok=None,
        # binning
        bin_ind=None,
        bin_dlamb=None,
        # unused (data)
        **kwdargs,
    ):

        # ---------------------
        # get lamb limits + iok

        # for pulses
        lamb00 = lamb[0]
        lambD = lamb[-1] - lamb[0]
        lambm = 0.5*(lamb[-1] + lamb[0])

        # iok
        if bin_iok is not None and bin_ind is False:
            lamb = lamb[bin_iok]

        # ----------
        # initialize

        shape = tuple(list(lamb.shape) + [x_free.size])
        val = np.zeros(shape, dtype=float)
        lamb = lamb[:, None]

        # -------------------
        # rescale

        if scales is not None:
            x_free = x_free * scales

        # ----------------------------
        # get x_full from constraints

        if c0 is None:
            x_full = x_free
        else:
            x_full = c2.dot(x_free**2) + c1.dot(x_free) + c0

        # -------
        # poly

        kfunc = 'poly'
        if dind.get(kfunc) is not None:

            lamb_rel = (lamb - lambm) / lambD

            vind = dind['jac'][kfunc].get('a0')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = 1. * scales[None, ival]

            vind = dind['jac'][kfunc].get('a1')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = lamb_rel * scales[None, ival]

            vind = dind['jac'][kfunc].get('a2')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = lamb_rel**2 * scales[None, ival]

        # --------
        # exp_lamb

        kfunc = 'exp_lamb'
        if dind.get(kfunc) is not None:
            amp = x_full[dind[kfunc]['amp']['ind']][None, :]
            rate = x_full[dind[kfunc]['rate']['ind']][None, :]

            exp_on_lamb = np.exp(- rate / lamb) / lamb

            vind = dind['jac'][kfunc].get('amp')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = exp_on_lamb[None, ivar] * scales[None, ival]

            vind = dind['jac'][kfunc].get('rate')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    - amp[:, ivar] * exp_on_lamb[:, ivar]
                    * scales[None, ival] / lamb
                )

        # -----------------
        # all gaussians

        kfunc = 'gauss'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][None, :]
            sigma = x_full[dind[kfunc]['sigma']['ind']][None, :]
            vccos = x_full[dind[kfunc]['vccos']['ind']][None, :]
            lamb0 = param_val[dind[kfunc]['lamb0']][None, :]

            dlamb = lamb - lamb0*(1 + vccos)
            exp = np.exp(-dlamb**2/(2*sigma**2))

            vind = dind['jac'][kfunc].get('amp')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = exp[:, ivar] * scales[None, ival]

            vind = dind['jac'][kfunc].get('vccos')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    amp[:, ivar] * exp[:, ivar]
                    * (dlamb[:, ivar] / sigma[:, ivar]**2) * lamb0[:, ivar]
                    * scales[None, ival]
                )

            vind = dind['jac'][kfunc].get('sigma')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    amp[:, ivar] * exp[:, ivar]
                    * (dlamb[:, ivar]**2 / sigma[:, ivar]**3)
                    * scales[None, ival]
                )

        # -------------------
        # all Lorentzians

        kfunc = 'lorentz'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][None, :]
            gam = x_full[dind[kfunc]['gam']['ind']][None, :]
            vccos = x_full[dind[kfunc]['vccos']['ind']][None, :]
            lamb0 = param_val[dind[kfunc]['lamb0']][None, :]

            # https://en.wikipedia.org/wiki/Cauchy_distribution
            # value at lamb0 = amp / (pi * gam)

            lamb_on_gam = (lamb - lamb0*(1 + vccos)) / gam

            vind = dind['jac'][kfunc].get('amp')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    scales[None, ival] / (1 + lamb_on_gam[:, ivar]**2)
                )

            vind = dind['jac'][kfunc].get('vccos')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    (amp[:, ivar] * lamb0[:, ivar] / gam[:, ivar])
                    * scales[None, ival]
                    * 2 * lamb_on_gam[:, ivar]
                    / (1 + lamb_on_gam[:, ivar]**2)**2
                )

            vind = dind['jac'][kfunc].get('gam')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    amp[:, ivar] * 2 * lamb_on_gam[:, ivar]**2
                    / (1 + lamb_on_gam[:, ivar]**2)**2
                    * scales[None, ival] / gam[:, ivar]
                )

        # -------------------
        # all pvoigt

        kfunc = 'pvoigt'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][None, :]
            sigma = x_full[dind[kfunc]['sigma']['ind']][None, :]
            gam = x_full[dind[kfunc]['gam']['ind']][None, :]
            vccos = x_full[dind[kfunc]['vccos']['ind']][None, :]
            lamb0 = param_val[dind[kfunc]['lamb0']][None, :]

            fg = 2 * np.sqrt(2*np.log(2)) * sigma
            fl = 2 * gam

            ftot_norm = (
                fg**5 + 2.69269*fg**4*fl + 2.42843*fg**3*fl**2
                + 4.47163*fg**2*fl**3 + 0.07842*fg*fl**4 + fl**5
            )
            ftot = ftot_norm ** (1./5.)
            ratio = fl / ftot

            # eta
            eta = 1.36603 * ratio - 0.47719 * ratio**2 + 0.11116 * ratio**3

            # update widths of gauss and Lorentz
            sigma2 = ftot / (2 * np.sqrt(2*np.log(2)))
            gam2 = ftot / 2.

            dlamb = lamb - lamb0*(1 + vccos)
            exp = np.exp(-dlamb**2/(2*sigma2**2))
            lamb_on_gam = dlamb / gam2
            lorentz_norm = 1. / (1 + lamb_on_gam**2)

            # weighted sum
            # ind = dind['func'][kfunc]['ind']
            # val[ind, ...] = amp * (
            #     eta / (1 + ((lamb - lamb0*(1 + vccos)) / gam)**2)
            #     + (1-eta) * np.exp(-(lamb - lamb0*(1 + vccos))**2/(2*sigma**2))
            # )

            vind = dind['jac'][kfunc].get('amp')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    (
                        eta[:, ivar] * lorentz_norm[:, ivar]
                        + (1-eta[:, ivar])*exp[:, ivar]) * scales[None, ival]
                )

            vind = dind['jac'][kfunc].get('vccos')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = amp[:, ivar] * scales[None, ival] * (
                    eta[:, ivar] * (lamb0[:, ivar] / gam2[:, ivar]) *
                    2 * lamb_on_gam[:, ivar] / (1 + lamb_on_gam[:, ivar]**2)**2
                    + (1 - eta[:, ivar]) * exp[:, ivar]
                    * (dlamb[:, ivar] / sigma2[:, ivar]**2) * lamb0[:, ivar]
                )

            # --------------
            # widths

            # sigma
            vind = dind['jac'][kfunc].get('sigma')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                ds_fg = 2 * np.sqrt(2*np.log(2))
                ds_ftot = (1/5) * ftot_norm**(-4./5.) * ds_fg * (
                    5*fg**4 + 2.69269*4*fg**3*fl + 2.42843*3*fg**2*fl**2
                    + 4.47163*2*fg*fl**3 + 0.07842*fl**4
                )
                ds_ratio = (-fl/ftot**2) * ds_ftot
                ds_eta = ds_ratio * (
                    1.36603 - 0.47719 * 2 * ratio + 0.11116 * 3 * ratio**2
                )
                ds_sigma2 = ds_ftot / (2 * np.sqrt(2*np.log(2)))
                ds_gam2 = ds_ftot / 2.

                ds_lamb_on_gam = - ds_gam2 * dlamb / gam2**2
                ds_exp = ds_sigma2 * (dlamb**2/sigma2**3) * exp

                val[:, ival] = amp[:, ivar] * scales[None, ival] * (
                    ds_eta[:, ivar] / (1 + lamb_on_gam[:, ivar]**2)
                    + eta[:, ivar] * ds_lamb_on_gam[:, ivar]
                    * (-2*lamb_on_gam[:, ivar])
                    / (1 + lamb_on_gam[:, ivar]**2)**2
                    - ds_eta[:, ivar] * exp[:, ivar]
                    + (1 - eta[:, ivar]) * ds_exp[:, ivar]
                )

            vind = dind['jac'][kfunc].get('gam')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                dg_fl = 2
                dg_ftot = (1/5) * ftot_norm**(-4./5.) * dg_fl * (
                    2.69269*fg**4 + 2.42843*fg**3*2*fl
                    + 4.47163*fg**2*3*fl**2 + 0.07842*fg*4*fl**3 + 5*fl**4
                )
                dg_ratio = dg_fl/ftot + (-fl/ftot**2) * dg_ftot
                dg_eta = dg_ratio * (
                    1.36603 - 0.47719 * 2 * ratio + 0.11116 * 3 * ratio**2
                )
                dg_sigma2 = dg_ftot / (2 * np.sqrt(2*np.log(2)))
                dg_gam2 = dg_ftot / 2.

                dg_lamb_on_gam = - dg_gam2 * dlamb / gam2**2
                dg_exp = dg_sigma2 * (dlamb**2/sigma2**3) * exp

                val[:, ival] = amp[:, ivar] * scales[None, ival] * (
                    dg_eta[:, ivar] / (1 + lamb_on_gam[:, ivar]**2)
                    + eta[:, ivar] * dg_lamb_on_gam[:, ivar]
                    * (-2*lamb_on_gam[:, ivar])
                    / (1 + lamb_on_gam[:, ivar]**2)**2
                    - dg_eta[:, ivar] * exp[:, ivar]
                    + (1 - eta[:, ivar]) * dg_exp[:, ivar]
                )

        # -------------------
        # all pulse_exp

        kfunc = 'pulse_exp'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][None, :]
            tau = x_full[dind[kfunc]['tau']['ind']][None, :]
            tup = x_full[dind[kfunc]['t_up']['ind']][None, :]
            tdown = x_full[dind[kfunc]['t_down']['ind']][None, :]

            ind0 = lamb >= (lamb00 + lambD * tau)
            dlamb = lamb - (lamb00 + lambD * tau)
            exp_up = np.exp(-dlamb/tup)
            exp_down = np.exp(-dlamb/tdown)

            # amp
            vind = dind['jac'][kfunc].get('amp')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    scales[None, ival]
                    * ind0[:, ivar] * (exp_down[:, ivar] - exp_up[:, ivar])
                )

            # tau
            vind = dind['jac'][kfunc].get('tau')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                dtau_exp_up = (
                    scales[ival]
                    * exp_up[:, ivar]
                    * (lambD/tup[:, ivar])
                )
                dtau_exp_down = (
                    scales[None, ival]
                    * exp_down[:, ivar]
                    * (lambD/tdown[:, ivar])
                )
                # dtau_ind0 = ind0_t / lambd
                val[:, ival] = amp[:, ivar] * (
                    ind0[:, ivar] * (dtau_exp_down - dtau_exp_up)
                    # + dtau_ind0 * (exp_down - exp_up)
                )

            # tup
            vind = dind['jac'][kfunc].get('t_up')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    amp[:, ivar]
                    * ind0[:, ivar] * scales[None, ival] * (
                        - exp_up[:, ivar] * (dlamb[:, ivar]/tup[:, ivar]**2)
                    )
                )

            # tdown
            vind = dind['jac'][kfunc].get('t_down')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    amp[:, ivar]
                    * ind0[:, ivar] * scales[None, ival] * (
                        exp_down[:, ivar] * (dlamb[:, ivar]/tdown[:, ivar]**2)
                    )
                )

        # -------------------
        # all pulse_gauss

        kfunc = 'pulse_gauss'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']][None, :]
            tau = x_full[dind[kfunc]['tau']['ind']][None, :]
            tup = x_full[dind[kfunc]['t_up']['ind']][None, :]
            tdown = x_full[dind[kfunc]['t_down']['ind']][None, :]

            indup = (lamb < (lamb00 + lambD * tau))
            inddown = (lamb >= (lamb00 + lambD * tau))

            dlamb = lamb - (lamb00 + lambD * tau)
            exp_up = np.exp(-dlamb**2/tup**2)
            exp_down = np.exp(-dlamb**2/tdown**2)

            # val[ind, ...] = (
            #     amp * (
            #         indup * np.exp(-(lamb - t0)**2/tup**2)
            #         + inddown * np.exp(-(lamb - t0)**2/tdown**2)
            #     )
            # )

            # amp
            vind = dind['jac'][kfunc].get('amp')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = scales[None, ival] * (
                    indup[:, ivar] * exp_up[:, ivar]
                    + inddown[:, ivar] * exp_down[:, ivar]
                )

            # tau
            vind = dind['jac'][kfunc].get('t0')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                dtau_exp_up = (
                    scales[None, ival]
                    * exp_up[:, ivar]
                    * (-1/tup[:, ivar]**2) * (-2*lambD*dlamb[:, ivar])
                )
                dtau_exp_down = (
                    scales[None, ival]
                    * exp_down[:, ivar]
                    * (-1/tdown[:, ivar]**2) * (-2*lambD*dlamb[:, ivar])
                )
                # dtau_ind0 = ind0_t / lambd
                val[:, ival] = amp[:, ivar] * (
                    indup[:, ivar] * dtau_exp_up
                    + inddown[:, ivar] * dtau_exp_down
                )

            # tup
            vind = dind['jac'][kfunc].get('t_up')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    amp[:, ivar] * indup[:, ivar] * scales[None, ival] * (
                        exp_up[:, ivar] * (2*dlamb[:, ivar]**2/tup[:, ivar]**3)
                    )
                )

            # tdown
            vind = dind['jac'][kfunc].get('t_down')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                val[:, ival] = (
                    amp[:, ivar] * inddown[:, ivar] * scales[None, ival] * (
                        exp_down[:, ivar] * (2*dlamb[:, ivar]**2/tdown[:, ivar]**3)
                    )
                )

        # -------------------
        # all lognorm

        kfunc = 'lognorm'
        if dind.get(kfunc) is not None:

            amp = x_full[dind[kfunc]['amp']['ind']]
            tau = x_full[dind[kfunc]['tau']['ind']]
            sigma = x_full[dind[kfunc]['sigma']['ind']]
            mu = x_full[dind[kfunc]['mu']['ind']]

            # max at t - t0 = exp(mu - sigma**2)
            # max = amp * exp(sigma**2/2 - mu)
            # variance = (exp(sigma**2) - 1) * exp(2mu + sigma**2)
            # skewness = (exp(sigma**2) + 2) * sqrt(exp(sigma**2) - 1)

            # amp
            vind = dind['jac'][kfunc].get('amp')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                for ii, i0 in enumerate(ival):
                    ioki = lamb[:, 0] > (lamb00 + lambD * tau[ivar[ii]])
                    dlamb = lamb[ioki, 0] - (lamb00 + lambD * tau[ivar[ii]])

                    log_mu = np.log(dlamb) - mu[ivar[ii]]
                    exp = np.exp(-(log_mu)**2 / (2.*sigma[ivar[ii]]**2))

                    val[ioki, i0] = (scales[i0] / dlamb) * exp

            # tau
            vind = dind['jac'][kfunc].get('tau')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                for ii, i0 in enumerate(ival):
                    ioki = lamb[:, 0] > (lamb00 + lambD * tau[ivar[ii]])
                    dlamb = lamb[ioki, 0] - (lamb00 + lambD * tau[ivar[ii]])

                    log_mu = np.log(dlamb) - mu[ivar[ii]]
                    exp = np.exp(-(log_mu)**2 / (2.*sigma[ivar[ii]]**2))

                    dtau_inv_dlamb = scales[i0] * lambD/dlamb**2
                    dtau_logmu = scales[i0] * (-lambD) / dlamb
                    dtau_exp = (
                        exp * dtau_logmu * (-2*log_mu / (2.*sigma[ivar[ii]]**2))
                    )

                    val[ioki, i0] = amp[ivar[ii]] * (
                        dtau_inv_dlamb * exp
                        + (1/dlamb) * dtau_exp
                    )

            # sigma
            vind = dind['jac'][kfunc].get('sigma')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                for ii, i0 in enumerate(ival):
                    ioki = lamb[:, 0] > (lamb00 + lambD * tau[ivar[ii]])
                    dlamb = lamb[ioki, 0] - (lamb00 + lambD * tau[ivar[ii]])

                    log_mu = np.log(dlamb) - mu[ivar[ii]]
                    exp = np.exp(-(log_mu)**2 / (2.*sigma[ivar[ii]]**2))

                    val[ioki, i0] = (
                        (amp[ivar[ii]] / dlamb) * exp * scales[i0]
                        * (log_mu**2/sigma[ivar[ii]]**3)
                    )

            # mu
            vind = dind['jac'][kfunc].get('mu')
            if vind is not None:
                ival, ivar = vind['val'], vind['var']
                for ii, i0 in enumerate(ival):
                    ioki = lamb[:, 0] > (lamb00 + lambD * tau[ivar[ii]])
                    dlamb = lamb[ioki, 0] - (lamb00 + lambD * tau[ivar[ii]])

                    log_mu = np.log(dlamb) - mu[ivar[ii]]
                    exp = np.exp(-(log_mu)**2 / (2.*sigma[ivar[ii]]**2))

                    val[ioki, i0] = 0.1 * (
                        (amp[ivar[ii]] / dlamb) * exp * scales[i0]
                        * (-1/(2.*sigma[ivar[ii]]**2)) * (-2*log_mu)
                    )

                # val[iok, i0] = (
                #     (amp[ii] / dlamb)
                #     * np.exp(-(np.log(dlamb) - mu[ii])**2 / (2.*sigma[ii]**2))
                # )

        # -------
        # binning

        if bin_ind is not False:
            val = np.add.reduceat(val * bin_dlamb[:, None], bin_ind, axis=0)
            if iok is not None:
                val = val[iok, :]

        return val

    return func


# #!/usr/bin/python3
# -*- coding: utf-8 -*-


import os


import numpy as np
import pandas as pd
import datastock as ds


_PATH_HERE = os.path.dirname(__file__)


#############################################
#############################################
#       add
#############################################


def add_ion(
    coll=None,
    key=None,
):

    # -------------------------
    # list of possible elements
    # -------------------------

    symbols, names, Z, A = _get_table()

    # ----------------
    # identify ion
    # ----------------

    # check if all ions should be added
    add_all = _check_key(key, symbols=symbols)

    # add all ions for given element
    if add_all is True:
        ions = [f"{key}{qq}+" for qq in range(0, Z[symbols.index(key)+1])]
        for ion in ions:
            coll.add_ion(ion)

    # add only specified ion
    else:

        # check element
        symbol = ''.join([ss for ss in key[:-1] if not ss.isnumeric()])
        symbol = ds._generic_check._check_var(
            symbol.capitalize(), 'symbol',
            types=str,
            allowed=symbols,
        )
        ind = symbols.index(symbol)

        # check charge
        charge = int(''.join([ss for ss in key[1:-1] if ss.isnumeric()]))
        charge = ds._generic_check._check_var(
            charge, 'charge',
            types=int,
            allowed=np.arange(0, Z[ind]+1),
        )

        # --------------
        # get proper key
        # --------------

        name = names[ind]
        z = Z[ind]
        aa = A[ind]

        # rebuild key
        key = f"{symbol}{charge}+"

        # ----------------
        # get iso-electronic sequence
        # ----------------

        isoelect = _get_isoelect(symbols=symbols, Z=Z, z=z, q=charge)

        # -------------
        # add if not in Collection
        # -------------

        if key not in coll.dobj.get(coll._which_ion, {}).keys():

            coll.add_obj(
                which=coll._which_ion,
                key=key,
                element=symbol,
                element_name=name,
                A=aa,
                Z=z,
                q=charge,
                isoelect=isoelect,
            )

    return


#############################################
#############################################
#       remove
#############################################


def remove_ion(coll=None, key=None, propagate=None):

    # -------------------------
    # list of possible elements
    # -------------------------

    symbols, names, Z, A = _get_table()

    # ----------------
    # identify ion
    # ----------------

    # check if all ions should be added
    add_all = _check_key(key, symbols=symbols)

    # add all ions for given element
    if add_all is True:
        ions = [f"{key}{qq}+" for qq in range(0, Z[symbols.index(key)+1])]
        for ion in ions:
            if ion in coll.dobj.get('ion', {}).keys():
                coll.remove_ion(ion, propagate=propagate)

    # add only specified ion
    else:
        if key in coll.dobj.get('ion', {}).keys():
            coll.remove_obj(which='ion', key=key, propagate=propagate)


#############################################
#############################################
#       check key
#############################################


def _call_err(key):
    msg = (
        "Arg key must be a str of the form 'LLXX+', where:\n"
        "\t- LL: is one or 2 letter\n"
        "\t- XX: is one or 2 integers\n"
        "Or it can be just an element 'LL' => all ions are added\n"
        f"Provided:\n\t{key}\n"
    )
    raise Exception(msg)


def _check_key(key, symbols=None):

    c0 = (
        isinstance(key, str)
        and (not key[0].isnumeric())
    )

    if not c0:
        _call_err(key)

    # only element name provided => add all ions
    if len(key) <= 2 and not any([ss.isnumeric() or ss == '+' for ss in key]):
        key = key.capitalize()
        if key not in symbols:
            _call_err(key)
        return True

    # otherwise => add single ion
    else:

        c0 = (
            len(key) >= 3
            and key.endswith('+')
            and len([ss.isnumeric() for ss in key[1:-1]]) > 0
        )

        if not c0:
            _call_err(key)
        return False


#############################################
#############################################
#       Database
#############################################


def _get_table():
    """ get symbols, names, atomic numbers (Z) and masses, sorted by Z """

    try:
        symbols, names, Z, A = _get_from_periodictable()

    except Exception as err:
        symbols, names, Z, A = _get_from_local()

    return symbols, names, Z, A


def _get_from_periodictable(update=False):
    """ Get elements characteristics from library periodictable
    """

    import periodictable as pt

    symbols = [str(el) for el in pt.elements]
    Z = [getattr(pt, el).number for el in symbols]
    A = [getattr(pt, el).mass for el in symbols]
    names = [getattr(pt, el).name for el in symbols]

    # sorting index (by atomic number)
    inds = np.argsort(Z)

    # sort
    symbols = tuple([symbols[ii] for ii in inds])
    names = tuple([names[ii].lower() for ii in inds])
    Z = tuple([Z[ii] for ii in inds])
    A = tuple([A[ii] for ii in inds])

    return symbols, names, Z, A


def _get_from_local():

    pfe = os.path.join(_PATH_HERE, '_class00_check_ions_table.csv')
    df = pd.read_csv(pfe)

    symbols = df.symbol.to_list()
    names = df.name.to_list()
    Z = df.Z.to_list()
    A = df.A.to_list()

    return symbols, names, Z, A


#############################################
#############################################
#       iso-electronic sequence
#############################################


def _get_isoelect(symbols=None, Z=None, z=None, q=None):

    k0 = symbols[Z.index(z-q)]
    return f"{k0}-like"



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 15 21:53:45 2024

@author: dvezinet
"""
# Built-in
import os


# Standard
import matplotlib.pyplot as plt


# spectrally-specific
from ._setup_teardown import setup_module0, teardown_module0
from .._class02_SpectralFit import SpectralFit as Collection
# from .._saveload import load
from .._class01_show import get_available_spectral_model_functions
from .._class01_display_models import display_spectral_model_function
from . import _spectralfit_input as _inputs


_PATH_HERE = os.path.dirname(__file__)
_PATH_INPUT = os.path.join(_PATH_HERE, 'input')
_PATH_OUTPUT = os.path.join(_PATH_HERE, 'output')


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    setup_module0(module)


def teardown_module(module):
    teardown_module0(module)


#######################################################
#
#     Instanciate and populate
#
#######################################################


class Test00_Populate():

    # ------------------------
    #   setup and teardown
    # ------------------------

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):

        # instanciate
        self.coll = Collection()

        # add data
        _inputs.add_data(self.coll)
        _inputs.add_xfree(self.coll)

        # add spectral lines
        pfe_json = os.path.join(_PATH_INPUT, 'spectrallines.json')
        self.coll.add_spectral_lines_from_file(pfe_json)

    # ------------------------
    #   Populating
    # ------------------------

    # -------------
    # add models

    def test00_get_available_spectral_model_funcstions(self):
        get_available_spectral_model_functions()

    def test01_display_spectral_model_function(self):
        lm = [
            'poly', 'exp_lamb',
            'gauss', 'lorentz', 'pvoigt',
            'pulse_exp', 'pulse_gauss', 'lognorm',
        ]
        for kk in lm:
            _ = display_spectral_model_function(kk)
        plt.close('all')

    def test02_add_spectral_model(self):
        _inputs.add_models(self.coll)

    def test03_get_spectral_model_func(self):
        _inputs.get_spectral_model_func(self.coll)

    def test04_interpolate_spectral_model(self):
        _inputs.interpolate_spectral_model(self.coll)

    def test05_get_spectral_model_moments(self):
        _inputs.get_spectral_model_moments(self.coll)

    def test06_plot_spectral_model(self):
        _inputs.plot_spectral_model(self.coll)


# -*- coding: utf-8 -*-
"""
The nist-compatibility module of tofu

"""


import traceback


_PKG = 'spectrally'


try:
    try:
        from spectrally.nist._requests import *
    except Exception:
        from ._requests import *

except Exception as err:
    msg = (
        str(traceback.format_exc())
        + f"\n\n\t=> optional sub-package {_PKG}.nist not usable\n"
    )
    raise Exception(msg)


del traceback, _PKG



# #!/usr/bin/python3
# -*- coding: utf-8 -*-


import copy


from ._class01_SpectralModel import SpectralModel as Previous
from . import _class02_check_fit as _check_fit
from . import _class02_show as _show
from . import _class02_binning as _binning
from . import _class02_compute_fit as _compute_fit
from . import _class02_plot_valid as _plot_valid
from . import _class02_plot_fit as _plot_fit


__all__ = ['SpectralFit']


#############################################
#############################################
#       Spectral Lines
#############################################


class SpectralFit(Previous):

    _ddef = copy.deepcopy(Previous._ddef)

    # _show_in_summary_core = ['shape', 'ref', 'group']
    _dshow = dict(Previous._dshow)

    _ddef['params']['dobj'] = {
    }

    # ###################
    # -------------------
    # Spectral models
    # -------------------

    def _get_show_obj(self, which=None):
        if which == self._which_fit:
            return _show._show
        else:
            return super()._get_show_obj(which)

    # ###################
    # -------------------
    # Spectral fits
    # -------------------

    # -------------------
    # add spectral fit
    # -------------------

    def add_spectral_fit(
        self,
        # keys
        key=None,
        key_model=None,
        key_data=None,
        key_sigma=None,
        absolute_sigma=None,
        # wavelength
        key_lamb=None,
        # optional 2d fit
        key_bs=None,
        key_bs_vect=None,
        # fit parameters
        dparams=None,
        dvalid=None,
        # compute options
        chain=None,
    ):

        _check_fit._check(
            coll=self,
            # keys
            key=key,
            key_model=key_model,
            key_data=key_data,
            key_sigma=key_sigma,
            absolute_sigma=absolute_sigma,
            # wavelength
            key_lamb=key_lamb,
            # optional 2d fit
            key_bs=key_bs,
            key_bs_vect=key_bs_vect,
            # fit parameters
            dparams=dparams,
            dvalid=dvalid,
            # compute options
            chain=chain,
        )

        return

    # -------------------
    # spectral fit binning
    # -------------------

    def get_spectral_fit_binning_dict(
        self,
        binning=None,
        lamb=None,
        iok=None,
        axis=None,
    ):
        """


        Returns
        -------
        dout: dict
            dict containing all useful parameters for binning

        """
        return _binning.main(
            coll=self,
            binning=binning,
            lamb=lamb,
            iok=iok,
            axis=axis,
        )

    # -------------------
    # compute spectral fit
    # -------------------

    def compute_spectral_fit(
        self,
        key=None,
        # binning
        binning=None,
        # solver options
        solver=None,
        dsolver_options=None,
        # options
        chain=None,
        dscales=None,
        dbounds_low=None,
        dbounds_up=None,
        dx0=None,
        # storing
        store=None,
        overwrite=None,
        # options
        strict=None,
        verb=None,
        timing=None,
    ):

        return _compute_fit.main(
            coll=self,
            key=key,
            # binning
            binning=binning,
            # solver options
            solver=solver,
            dsolver_options=dsolver_options,
            # options
            chain=chain,
            dscales=dscales,
            dbounds_low=dbounds_low,
            dbounds_up=dbounds_up,
            dx0=dx0,
            # storing
            store=store,
            overwrite=overwrite,
            # options
            strict=strict,
            verb=verb,
            timing=timing,
        )

    # ----------------------------------
    # plot spectral fit data validity
    # ----------------------------------

    def plot_spectral_fit_input_validity(
        self,
        key=None,
        keyY=None,
        dref_vectorY=None,
        # options
        dprop=None,
        vmin=None,
        vmax=None,
        cmap=None,
        plot_text=None,
        # figure
        dax=None,
        fs=None,
        dmargin=None,
        tit=None,
        # interactivity
        connect=True,
        dinc=None,
        show_commands=None,
    ):
        """ Plot the validity map if input data


        Parameters
        ----------
        key : str, optional
            DESCRIPTION. The default is None.
        keyY : str, optional
            DESCRIPTION. The default is None.
        dref_vectorY : dict, optional
            DESCRIPTION. The default is None.
        dprop : TYPE, optional
            DESCRIPTION. The default is None.
        vmin : float, optional
            DESCRIPTION. The default is None.
        vmax : float, optional
            DESCRIPTION. The default is None.
        cmap : str, optional
            DESCRIPTION. The default is None.
        plot_text : bool, optional
            DESCRIPTION. The default is None.
        dax : dict, optional
            DESCRIPTION. The default is None.
        fs : tuple, optional
            DESCRIPTION. The default is None.
        dmargin : dict, optional
            DESCRIPTION. The default is None.
        tit : str, optional
            DESCRIPTION. The default is None.
        connect : bool, optional
            DESCRIPTION. The default is True.
        dinc : dict, optional
            DESCRIPTION. The default is None.
        show_commands : bool, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        dax:    Collection
            collection of axes

        """

        return _plot_valid.plot(
            coll=self,
            key=key,
            keyY=keyY,
            dref_vectorY=dref_vectorY,
            # options
            dprop=dprop,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            plot_text=plot_text,
            # figure
            dax=dax,
            fs=fs,
            dmargin=dmargin,
            tit=tit,
            # interactivity
            connect=connect,
            dinc=dinc,
            show_commands=show_commands,
        )

    # ----------------------------
    # plot spectral fit
    # ----------------------------

    def plot_spectral_fit(
        self,
        key=None,
        keyY=None,
        # options
        details=None,
        # uncertainty propagation
        uncertainty_method=None,
        # plotting
        dprop=None,
        vmin=None,
        vmax=None,
        # lines labels
        lines_labels=True,
        lines_labels_color=None,
        lines_labels_rotation=None,
        lines_labels_horizontalalignment=None,
        # figure
        dax=None,
        fs=None,
        dmargin=None,
        tit=None,
        # interactivity
        nmax=None,
        connect=None,
        dinc=None,
        show_commands=None,
        # timing
        timing=None,
    ):

        """ Plot a spectral model using specified data

        lamb can be:
            - a key to an existing vector
            - a user-provided vector (1d np.ndarray)

        """

        return _plot_fit.main(
            coll=self,
            key=key,
            keyY=keyY,
            # options
            details=details,
            # uncertainty propagation
            uncertainty_method=uncertainty_method,
            # plotting
            dprop=dprop,
            vmin=vmin,
            vmax=vmax,
            # lines labels
            lines_labels=lines_labels,
            lines_labels_color=lines_labels_color,
            lines_labels_rotation=lines_labels_rotation,
            lines_labels_horizontalalignment=lines_labels_horizontalalignment,
            # figure
            dax=dax,
            fs=fs,
            dmargin=dmargin,
            tit=tit,
            # interactivity
            nmax=nmax,
            connect=connect,
            dinc=dinc,
            show_commands=show_commands,
            # timing
            timing=timing,
        )


# #!/usr/bin/python3
# -*- coding: utf-8 -*-


import warnings


from .version import __version__
from ._class02_SpectralFit import SpectralFit as Collection
from ._saveload import load
from ._class01_show import *
from ._class01_display_models import *
from . import tests
from . import tutorials


_PKG = 'spectrally'


# -------------------------------------
#   Try importing optional subpackages
# -------------------------------------

msg = None
dsub = dict.fromkeys([
    'openadas', 'nist'
])

for sub in dsub.keys():
    try:
        exec(f'import {_PKG}.{sub} as {sub}')
        dsub[sub] = True

    except Exception as err:
        dsub[sub] = str(err)

# -------------------------------------
# If any error, populate warning and store error message
# -------------------------------------

lsubout = [sub for sub in dsub.keys() if dsub[sub] is not True]
if len(lsubout) > 0:
    lsubout = [f'{_PKG}.{ss}' for ss in lsubout]
    msg = (
        "\nThe following subpackages are not available:\n"
        + "\n".join([f"\t- {sub}" for sub in lsubout])
        + f"\n  => see print({_PKG}.dsub[<subpackage>]) for details."
    )
    warnings.warn(msg)

# -------------------------------------
# Add optional subpackages to __all__
# -------------------------------------

__all__ = []
for sub in dsub.keys():
    if dsub[sub] is True:
        __all__.append(sub)


# clean-up the mess
del warnings, lsubout, sub, msg, _PKG


#!/usr/bin/env python

# Built-in
import sys
import os
from shutil import copyfile


_HERE = os.path.abspath(os.path.dirname(__file__))


# import parser dict
sys.path.insert(1, _HERE)
from _dparser import _DPARSER
_ = sys.path.pop(1)


_PKG_NAME = 'spectrally'


###################################################
###################################################
#        Main
###################################################


def main():
    # Parse input arguments
    # Instanciate parser
    ddef, parser = _DPARSER['custom']()

    # Parse arguments
    args = parser.parse_args()

    # Call function
    custom(ddef=ddef, **dict(args._get_kwargs()))

    return


###################################################
###################################################
#       function
###################################################


def custom(
    target=None,
    source=None,
    files=None,
    directories=None,
    ddef=None,
):

    # --------------
    # Check inputs
    # --------------

    # get input
    kwd = locals()
    for k0 in set(ddef.keys()).intersection(kwd.keys()):
        if kwd[k0] is None:
            kwd[k0] = ddef[k0]

    # extract parameters
    target, source = kwd['target'], kwd['source']
    files, directories = kwd['files'], kwd['directories']

    # Caveat (up to now only relevant for _TARGET)
    if target != ddef['target']:
        msg = ""
        raise Exception(msg)

    # --------------
    # Check files
    # --------------

    if isinstance(files, str):
        files = [files]

    c0 = (
        not isinstance(files, list)
        or any([ff not in ddef['files'] for ff in files])
    )
    if c0 is True:
        msg = "All files should be in {}".format(ddef['files'])
        raise Exception(msg)

    # -------------------------------------------
    # Try creating directory and copying modules
    # -------------------------------------------

    try:
        # Create .spectrally/ if non-existent
        if not os.path.isdir(target):
            os.mkdir(target)

        # Create directories
        for dd in directories:
            if not os.path.isdir(os.path.join(target, dd)):
                os.mkdir(os.path.join(target, dd))

        # Copy files
        for ff in files:
            mod, f0 = ff.split('_')[1:]
            copyfile(
                os.path.join(source, mod, f'_{f0}'),
                os.path.join(target, ff),
            )

        msg = (
            f"A local copy of default {_PKG_NAME} parameters is now in:\n"
            f"\t{target}/\n"
            f"You can edit it to spice up your {_PKG_NAME}"
        )
        print(msg)

    except Exception as err:
        msg = (
            str(err)
            + "\n\nA problem occured\n"
            f"{_PKG_NAME}-custom failed to create a dir .{_PKG_NAME}/ in "
            f"your home {target}\n"
            "But it could not, check the error message above to debug\n"
            "Most frequent cause is a permission issue"
        )
        raise Exception(msg)

    return


###################################################
###################################################
#        __main__
###################################################


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Created on Tue Feb 20 14:44:51 2024

@author: dvezinet
"""

# #!/usr/bin/python3
# -*- coding: utf-8 -*-


import warnings


import numpy as np
import datastock as ds


# local
from . import _class02_compute_fit_1d as _compute_fit_1d


#############################################
#############################################
#       DEFAULTS
#############################################


#############################################
#############################################
#       Main
#############################################


def main(
    coll=None,
    key=None,
    # binning
    binning=None,
    # solver options
    solver=None,
    dsolver_options=None,
    # options
    chain=None,
    dscales=None,
    dbounds_low=None,
    dbounds_up=None,
    dx0=None,
    # storing
    store=None,
    overwrite=None,
    # options
    strict=None,
    verb=None,
    timing=None,
):
    """ Compute the fit of any previously added spectral fit

    """

    # ------------
    # check inputs
    # ------------

    (
        key, is1d,
        key_model,
        key_data, key_lamb,
        ref_data, ref_lamb,
        key_sigma,
        ref_cov, shape_cov,
        lamb, data, sigma, axis,
        absolute_sigma,
        chain,
        store, overwrite,
        strict, verb, verb_scp, timing,
    ) = _check(
        coll=coll,
        key=key,
        # options
        chain=chain,
        dscales=dscales,
        dbounds_low=dbounds_low,
        dbounds_up=dbounds_up,
        dx0=dx0,
        # storing
        store=store,
        overwrite=overwrite,
        # options
        strict=strict,
        verb=verb,
        timing=timing,
    )

    # ----------------
    # particular case

    ravel = False
    if data.ndim == 1:
        # don't change axis !
        data = data[:, None]
        sigma = sigma[:, None]
        shape_cov = shape_cov + (1,)
        ravel = True

    # ------------
    # solver_options
    # ------------

    solver, dsolver_options = _get_solver_options(
        solver=solver,
        dsolver_options=dsolver_options,
        verb_scp=verb_scp,
    )

    # ------------
    # fit
    # ------------

    if is1d is True:
        compute = _compute_fit_1d.main

    else:
        pass

    # -----------
    # verb

    if verb >= 1:
        msg = (
            "\n\n-------------------------------------------\n"
            f"\tComputing spectral fit '{key}'\n"
        )
        print(msg)

    # --------
    # run

    dout = compute(
        coll=coll,
        # keys
        key=key,
        key_model=key_model,
        key_data=key_data,
        key_lamb=key_lamb,
        # covariance
        ref_cov=ref_cov,
        shape_cov=shape_cov,
        # lamb, data, axis
        lamb=lamb,
        data=data,
        sigma=sigma,
        axis=axis,
        ravel=ravel,
        # binning
        binning=binning,
        # options
        chain=chain,
        dscales=dscales,
        dbounds_low=dbounds_low,
        dbounds_up=dbounds_up,
        dx0=dx0,
        # solver options
        solver=solver,
        dsolver_options=dsolver_options,
        absolute_sigma=absolute_sigma,
        # options
        strict=strict,
        verb=verb,
        verb_scp=verb_scp,
        timing=timing,
    )

    if dout is None:
        msg = (
            "No valid spectrum in chosen data:\n"
            f"\t- spectral_fit: {key}\n"
            f"\t- key_data: {key_data}\n"
        )
        warnings.warn(msg)
        return

    # --------------
    # format output
    # --------------

    if ravel is True:
        pass

    # ------------
    # store
    # ------------

    if store is True:

        _store(
            coll=coll,
            overwrite=overwrite,
            # keys
            key=key,
            key_model=key_model,
            key_data=key_data,
            key_lamb=key_lamb,
            # flags
            axis=axis,
            ravel=ravel,
            # dout
            dout=dout,
            # verb
            verb=verb,
        )

        dout = None

    return dout


#############################################
#############################################
#       Check
#############################################


def _check(
    coll=None,
    key=None,
    # options
    chain=None,
    dscales=None,
    dbounds_low=None,
    dbounds_up=None,
    dx0=None,
    # storing
    store=None,
    overwrite=None,
    # options
    strict=None,
    verb=None,
    timing=None,
):

    # --------------
    # key
    # --------------

    wsf = coll._which_fit
    lok_1d = [
        k0 for k0, v0 in coll.dobj.get(wsf, {}).items()
        if v0['key_bs'] is None
    ]
    lok_2d = [
        k0 for k0, v0 in coll.dobj.get(wsf, {}).items()
        if v0['key_bs'] is not None
    ]
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok_1d + lok_2d,
    )

    # is1d
    is1d = key in lok_1d

    if not is1d:
        msg = "2d fit not implemented yet"
        raise NotImplementedError(msg)

    # -----------
    # derive keys

    key_model = coll.dobj[wsf][key]['key_model']
    key_data = coll.dobj[wsf][key]['key_data']
    key_lamb = coll.dobj[wsf][key]['key_lamb']
    key_sigma = coll.dobj[wsf][key]['key_sigma']
    absolute_sigma = coll.dobj[wsf][key]['absolute_sigma']
    lamb = coll.ddata[key_lamb]['data']
    data = coll.ddata[key_data]['data']
    ref_data = coll.ddata[key_data]['ref']
    ref_lamb = coll.ddata[key_lamb]['ref']
    axis = ref_data.index(ref_lamb[0])

    # -----------
    # derive ref covariance

    # ref
    wsm = coll._which_model
    ref_nxfree = coll.dobj[wsm][key_model]['ref_nx']
    ref_cov = (
        ref_data[:axis]
        + (ref_nxfree, ref_nxfree)
        + ref_data[axis+1:]
    )
    # hence
    # ref_cov_axis = (axis, axis+1)

    # shape
    nxfree = coll.dref[ref_nxfree]['size']
    shape_cov = tuple([
        coll.dref[rr]['size'] for rr in ref_cov
    ])

    # ---------------------------------
    # voigt not handled for fitting yet
    # ---------------------------------

    lvoigt = [
        k0 for k0, v0 in coll.dobj[wsm][key_model]['dmodel'].items()
        if v0['type'] == 'voigt'
    ]
    if len(lvoigt) > 0:
        msg = (
            "Fitting is not implemented for voigt profiles yet\n"
            f"\t- key_model: `{key_model}`\n"
            f"\t- voigt functions: {lvoigt}\n"
            "Consider using another spectral model with pvoigt instead"
        )
        raise Exception(msg)

    # ---------------
    # sigma
    # ---------------

    if key_sigma == 'poisson':
        sigma = np.sqrt(np.abs(data))
        sigma[sigma == 0] = np.nanmin(sigma[sigma>0])

    elif isinstance(key_sigma, float):
        sigma = np.full(data.shape, key_sigma, dtype=float)

    elif key_sigma.endswith('%'):
        sigma = np.abs(data) * float(key_sigma[:-1]) / 100.
        sigma[sigma == 0] = np.nanmin(sigma[sigma>0])

    else:
        sigma = coll.ddata[key_sigma]['data']
        if sigma.ndim < data.ndim:
            resh = [sigma.size if ii == axis else 1 for ii in range(data.ndim)]
            sigma = sigma.reshape(resh)
            for ii, ss in enumerate(data.shape):
                if ii != axis:
                    sigma = np.repeat(sigma, ss, axis=ii)

    # --------------
    # chain
    # --------------

    chain = ds._generic_check._check_var(
        chain, 'chain',
        types=bool,
        default=False,
    )

    # --------------
    # store
    # --------------

    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=True,
    )

    # --------------
    # overwrite
    # --------------

    overwrite = ds._generic_check._check_var(
        overwrite, 'overwrite',
        types=bool,
        default=False,
    )

    if overwrite is True:
        ksol = coll.dobj[wsf][key]['key_sol']
        overwrite &= (ksol is not None and ksol in coll.ddata.keys())

    # --------------
    # strict
    # --------------

    strict = ds._generic_check._check_var(
        strict, 'strict',
        types=bool,
        default=False,
    )

    # --------------
    # verb
    # --------------

    def_verb = 2
    verb = ds._generic_check._check_var(
        verb, 'verb',
        default=def_verb,
        allowed=[False, True, 0, 1, 2, 3],
    )
    if verb is True:
        verb = def_verb
    elif verb is False:
        verb = 0

    # verb_scp
    verb_scp = {
        0: 0,
        1: 0,
        2: 0,
        3: 2,
    }[verb]

    # --------------
    # timing
    # --------------

    timing = ds._generic_check._check_var(
        timing, 'timing',
        types=bool,
        default=False,
    )

    return (
        key, is1d,
        key_model,
        key_data, key_lamb,
        ref_data, ref_lamb,
        key_sigma,
        ref_cov, shape_cov,
        lamb, data, sigma, axis,
        absolute_sigma,
        chain,
        store, overwrite,
        strict, verb, verb_scp, timing,
    )


#############################################
#############################################
#       Solver options
#############################################


def _get_solver_options(
    solver=None,
    dsolver_options=None,
    verb_scp=None,
):

    # -------------------
    # available solvers
    # -------------------

    lok = ['scipy.least_squares', 'scipy.curve_fit']
    solver = ds._generic_check._check_var(
        solver, 'solver',
        types=str,
        # default='scipy.least_squares',
        default='scipy.curve_fit',
        allowed=lok,
    )

    # -------------------
    # get default
    # -------------------

    if solver == 'scipy.least_squares':

        ddef = dict(
            # solver options
            method='trf',
            xtol=None,
            ftol=1e-10,
            gtol=None,
            tr_solver='exact',
            tr_options={},
            diff_step=None,
            max_nfev=6000,
            loss='linear',
            verbose=verb_scp,
        )

    elif solver == 'scipy.curve_fit':

        ddef = dict(
            # solver options
            full_output=True,
            # nan_policy='raise',  # not available on MacOS ?
            # common with least_squares
            method='trf',
            xtol=None,
            ftol=1e-10,
            gtol=None,
            tr_solver='exact',
            tr_options={},
            diff_step=None,
            max_nfev=6000,
            loss='linear',
            verbose=verb_scp,
        )

    else:
        raise NotImplementedError()

    # -------------------
    # implement
    # -------------------

    if dsolver_options is None:
        dsolver_options = {}

    if not isinstance(dsolver_options, dict):
        msg = (
        )
        raise Exception(msg)

    # add default values
    for k0, v0 in ddef.items():
        if dsolver_options.get(k0) is None:
            dsolver_options[k0] = v0

    # clear irrelevant keys
    lkout = [k0 for k0 in dsolver_options.keys() if k0 not in ddef.keys()]
    for k0 in lkout:
        del dsolver_options[k0]

    return solver, dsolver_options


#############################################
#############################################
#       store
#############################################


def _store(
    coll=None,
    overwrite=None,
    # keys
    key=None,
    key_model=None,
    key_data=None,
    key_lamb=None,
    # flags
    axis=None,
    ravel=None,
    # dout
    dout=None,
    # verb
    verb=None,
):

    # ------------
    # add ref
    # ------------

    wsm = coll._which_model
    wsf = coll._which_fit
    refx_free = coll.dobj[wsm][key_model]['ref_nx']
    ref = list(coll.ddata[key_data]['ref'])
    ref[axis] = refx_free

    ref_reduced = tuple([rr for ii, rr in enumerate(ref) if ii != axis])

    # ------------
    # add data
    # ------------

    # solution names
    ksol = f"{key}_sol"

    # solution arrays
    sol = dout['sol'].squeeze() if ravel is True else dout['sol']
    if dout['cov'] is None:
        kcov = None
        cov = None
    else:
        kcov = f"{key}_cov"
        cov = dout['cov'].squeeze() if ravel is True else dout['cov']

    # add as data
    if ksol in coll.ddata.keys():
        if overwrite is True:
            c0 = (
                coll.ddata[ksol]['ref'] == tuple(ref)
                and coll.dobj[wsf][key]['key_sol'] == ksol
            )
            if c0 is True:
                coll._ddata[ksol]['data'] = sol
                if cov is not None:
                    coll._ddata[kcov]['data'] = cov

            else:
                msg = (
                    "Fit sol '{ksol}' exists in ddata but not in spect_fit!"
                )
                raise Exception(msg)

        else:
            msg = (
                f"Aborting storing, spect_fit '{key}' was already computed\n"
                "Use overwrite=True to overwrite the former solution\n"
            )
            raise Exception(msg)

    else:
        # solution
        coll.add_data(
            key=ksol,
            data=sol,
            ref=tuple(ref),
            units=coll.ddata[key_data]['units'],
            dim='fit_sol',
        )

        # cov
        if cov is not None:
            cunits = coll.ddata[key_data]['units']
            if isinstance(cunits, str):
                cunits = f"{cunits}^2"
            else:
                cunits = cunits * cunits

            coll.add_data(
                key=kcov,
                data=cov,
                ref=dout['ref_cov'],
                units=cunits,
                dim='fit_cov',
            )

    # -------------
    # other outputs

    lk = [
        'cost', 'chi2n', 'time', 'status', 'nfev',
        'msg', 'validity', 'errmsg',
    ]
    dk_out = {k0: f"{key}_{k0}" for k0 in lk}

    if ravel is False:

        if overwrite is True:
            for k0, k1 in dk_out.items():
                coll._ddata[k1]['data'] = dout[k0]

        else:
            for k0, k1 in dk_out.items():
                coll.add_data(
                    key=k1,
                    data=dout[k0],
                    ref=ref_reduced,
                    units='',
                    dim='fit_out',
                )

    # ------------
    # store in fit
    # ------------

    wsf = coll._which_fit
    coll._dobj[wsf][key]['key_sol'] = ksol
    coll._dobj[wsf][key]['key_cov'] = kcov

    # scale, bounds, x0
    coll._dobj[wsf][key]['dinternal'] = {
        'scales': dout['scales'],
        'bounds0': dout['bounds0'],
        'bounds1': dout['bounds1'],
        'x0': dout['x0'],
        'binning': (
            False if dout['dbinning'] is False
            else dout['dbinning']['binning']
        ),
    }

    # solver output
    coll._dobj[wsf][key]['dsolver'] = {
        'solver': dout['solver'],
        'dsolver_options': dout['dsolver_options'],
    }

    if ravel is True:
        for k0, k1 in dk_out.items():
            coll._dobj[wsf][key]['dsolver'][k0] = dout[k0]

    else:
        for k0, k1 in dk_out.items():
            coll._dobj[wsf][key]['dsolver'][k0] = k1

    # ---------
    # verb
    # ----------

    if verb >= 2:
        msg = f"Storing under key '{ksol}'"
        print(msg)

    return



# Built-in
import os
import re
import itertools as itt
import warnings

# Common
import numpy as np
from scipy.interpolate import RectBivariateSpline as scpRectSpl


__all__ = ['step03_read', 'step03_read_all']


_PKG = 'spectrally'
_DTYPES = {
    'adf11': ['acd', 'ccd', 'scd', 'plt', 'prb'],
    'adf15': None,
}
_DEG = 1
_PECASFUNC = True


# #############################################################################
# #############################################################################
#                       Utility functions
# #############################################################################


def _get_PATH_LOCAL():
    pfe = os.path.join(os.path.expanduser('~'), f'.{_PKG}', 'openadas')
    if os.path.isdir(pfe):
        return pfe
    else:
        return None


def _get_subdir_from_pattern(path, pattern, mult=None):
    """ Get list of files matching patterns in path

    If no match => Exception
    If multiple matches
        => mult = True: pass
        => mult = 'warn': warning
        => mult = 'err': Exception
    """

    # -------------
    # Check inputs
    # -------------

    if mult is None:
        mult = 'err'
    if mult not in [True, 'warn', 'err']:
        msg = (
            "Arg mult must be in [True, 'warn', 'err']!\n"
            + "\t- provided: {}".format(mult)
        )
        raise Exception(msg)

    if isinstance(pattern, str):
        pattern = [pattern]

    # -------------
    # Get matches
    # -------------

    ld = [
        dd for dd in os.listdir(path)
        if os.path.isdir(os.path.join(path, dd))
        and all([pp in dd for pp in pattern])
    ]
    if len(ld) != 1:
        msg = (
            "You have no / many directories in your local "
            f"~/.{_PKG}/openadas/ matching the desired file type:\n"
            f"\t- path: {path}\n"
            f"\t- provided (all): {pattern}\n"
            f"\t- available: {ld}\n"
            f" => download the data with sp.openadas.step02_download()\n"
        )
        if len(ld) == 0:
            raise Exception(msg)
        else:
            if mult == 'err':
                raise Exception(msg)
            elif mult == 'warn':
                warnings.warn(msg)

    return [os.path.join(path, dd) for dd in ld]


def _get_available_elements_from_path(path=None, typ1=None):

    # ------------
    # Check inputs
    # ------------

    if not os.path.isdir(path):
        msg = (
            "Provided path is not an existing directory!\n"
            + "\t- provided: {}".format(path)
        )
        raise Exception(msg)

    # ----------------
    # check data types
    # ----------------

    ltyp = ['adf15']
    if typ1 not in ltyp:
        msg = (
            "Only the following types of files are handled up to now:\n"
            + "\t- handled: {}\n".format(ltyp)
            + "\t- provided: {}".format(typ1)
        )
        raise Exception(msg)

    # --------------
    # ADF15
    # --------------

    if typ1 == 'adf15':
        lf = [
            ff for ff in os.listdir(path)
            if all([ss in ff for ss in ['pec', '][']])
        ]
        element = [ff[ff.index('][')+2:] for ff in lf]

    return element


def _format_for_DataStock_adf15(
    dout,
    dsource0=None,
    dref0=None,
    ddata0=None,
    dlines0=None,
):
    """
    Format dout from step03_read_all() for SpectralLines object
    (separated te, ne, ions, sources, lines)
    """

    # ----------------------------------------------
    # Remove already known lines of dlines0 provided
    # ----------------------------------------------

    if dlines0 is not None:

        # --------------------------------------
        # Check for redundancy vs existing lines

        dk0 = {
            k0: [
                k1 for k1, v1 in dlines0.items()
                if np.sum([(
                    v1['ion'] == v0['ion']
                    and v1['source'] == v0['source']
                    and v1['transition'] == v0['transition'],
                    v1['ion'] == v0['ion']
                    and v1['source'] == v0['source']
                    and v1['symbol'] == v0['symbol']
                )]) == 1
            ]
            for k0, v0 in dout.items()
        }

        dk0 = {k0: v0 for k0, v0 in dk0.items() if len(v0) > 0}

        # -----------------------
        # raise warning if needed

        if len(dk0) > 0:
            msg = (
                "\nPossible error in openadas detected,\n"
                + "the following lines have same ion and transition but "
                + "different symbol (typ0typ1-isoel):\n"
                + "\n".join([
                    "\t- {}: {}".format(k0, v0) for k0, v0 in dk0.items()
                ])
                + "\n\n  => There might be redundancy / errors in openadas"
            )
            warnings.warn(msg)

        # -----------------
        # re-assemble dout

        dout = {
            k0: v0 for k0, v0 in dout.items()
            if not any([
                v1['ion'] == v0['ion']
                and v1['source'] == v0['source']
                and v1['transition'] == v0['transition']
                for v1 in dlines0.values()
            ])
        }

    # --------------------------
    # Get dict of unique sources
    # --------------------------

    lsource = sorted(set([v0['source'] for v0 in dout.values()]))
    dsource = {}
    if dsource0 is None:
        dsource = {
            'oa-adf15-{:02}'.format(ii): {'long': ss}
            for ii, ss in enumerate(lsource)
        }

    else:
        # Check against existing sources
        lsoa = [k0 for k0 in dsource0.keys() if 'oa-adf15' in k0]
        if len(lsoa) > 0:
            nmax = int(np.max([int(k0.split('-')[-1]) for k0 in lsoa])) + 1
        else:
            nmax = 0

        for ii, ss in enumerate(lsource):
            lk0 = [k0 for k0, v0 in dsource0.items() if v0.get('long') == ss]
            if len(lk0) == 0:
                k0 = f'oa-adf15-{nmax:02}'
                nmax += 1
            elif len(lk0) == 1:
                k0 = lk0[0]
            else:
                msg = f"\nMultiple possible matches for source {ss}"
                raise Exception(msg)

            dsource[k0] = {'long': ss}

    # ----------------------------
    # Get dict of unique Te and ne
    # ----------------------------

    # initialize dict and numbers
    dte, dne = {}, {}
    if dref0 is None:
        ite, ine = 0, 0

    else:
        ite = int(np.max([
            int(k0.split('-')[-1]) for k0 in dref0.keys()
            if 'nTe' in k0 and 'oa-adf15' in ddata0[k0]['source']
        ])) + 1
        ine = int(np.max([
            int(k0.split('-')[-1]) for k0 in dref0.keys()
            if 'nne' in k0 and 'oa-adf15' in ddata0[k0]['source']
        ])) + 1

    # scan
    for k0, v0 in dout.items():

        # Get source
        sour = [
            k1 for k1, v1 in dsource.items() if v1['long'] == v0['source']
        ][0]

        # fill dte
        kte = [
            kk for kk, vv in dte.items()
            if v0['te'].shape == vv['data'].shape
            and np.allclose(v0['te'], vv['data'])
            # and sour == vv['source']
        ]
        normal = dref0 is None
        if normal is False:
            # Check vs existing Te
            lk0 = [
                k1 for k1, v1 in dref0.items()
                if v0['te'].shape == ddata0[k1]['data'].shape
                and np.allclose(v0['te'], ddata0[k1]['data'])
                # and ddata0[k1]['source'] == sour
            ]
            if len(lk0) == 0:
                normal = True
            elif len(lk0) == 1:
                keyte = lk0[0]
                # dte[keyte] = {
                    # 'data': ddata0[lk0[0]]['data'],
                    # 'units': v0['te_units'],
                    # 'source': sour,
                    # 'dim': 'temperature',
                    # 'quant': 'Te',
                    # 'name': 'Te',
                # }
            elif len(lk0) > 1:
                msg = (
                    f"Multiple matches for dout['{k0}'] in dref0:\n"
                    f"\t- {lk0}"
                )
                raise Exception(msg)

        if normal is True:
            if len(kte) == 0:
                keyte = f'Te{ite:02}'
                dte[keyte] = {
                    'data': v0['te'],
                    'units': v0['te_units'],
                    'source': sour,
                    'dim': 'temperature',
                    'quant': 'Te',
                    'name': 'Te',
                }
                ite += 1
            elif len(kte) == 1:
                pass
            else:
                msg = (
                    "len(kte) != 1:\n"
                    + "\t- kte = {}\n".format(kte)
                )
                raise Exception(msg)

        dout[k0]['keyte'] = keyte

        # fill dne
        kne = [
            kk for kk, vv in dne.items()
            if v0['ne'].shape == vv['data'].shape
            and np.allclose(v0['ne'], vv['data'])
            # and sour == vv['source']
        ]
        normal = dref0 is None
        if normal is False:
            # Check vs existing ne
            lk0 = [
                k1 for k1, v1 in dref0.items()
                if v0['ne'].shape == ddata0[k1]['data'].shape
                and np.allclose(v0['ne'], ddata0[k1]['data'])
                # and ddata0[k1]['source'] == sour
            ]
            if len(lk0) == 0:
                normal = True
            elif len(lk0) == 1:
                keyne = lk0[0]
                # dne[keyne] = {
                    # 'data': ddata0[lk0[0]]['data'],
                    # 'units': v0['ne_units'],
                    # 'source': sour,
                    # 'dim': 'density',
                    # 'quant': 'ne',
                    # 'name': 'ne',
                # }
            elif len(lk0) > 1:
                msg = (
                    f"Multiple matches for dout['{k0}'] in dref0:\n"
                    f"\t- {k0}"
                )
                raise Exception(msg)

        if normal is True:
            if len(kne) == 0:
                keyne = f'ne{ine:02}'
                dne[keyne] = {
                    'data': v0['ne'],
                    'units': v0['ne_units'],
                    'source': sour,
                    'dim': 'density',
                    'quant': 'ne',
                    'name': 'ne',
                }
                ine += 1
            elif len(kne) == 1:
                pass
            else:
                msg = (
                    "len(kne) != 1:\n"
                    + "\t- kne = {}\n".format(kne)
                )
                raise Exception(msg)
        dout[k0]['keyne'] = keyne

    # ---------------
    # Get dict of pec
    # ---------------

    dpec = {
        f'{k0}_pec': {
            'data': v0['pec'], 'units': v0['pec_units'],
            'ref': (v0['keyne'], v0['keyte']),
            'source': [
                k1 for k1, v1 in dsource.items()
                if v1['long'] == v0['source']
            ][0],
            'dim': '<sigma v>',
            'quant': 'pec',
        }
        for k0, v0 in dout.items()
    }

    # --------
    # dlines
    # --------

    inds = np.argsort([v0['lambda0'] for v0 in dout.values()])
    lk0 = np.array(list(dout.keys()), dtype=str)[inds]
    dlines = {
        k0: {
            'ion': dout[k0]['ion'],
            'source': [
                k1 for k1, v1 in dsource.items()
                if v1['long'] == dout[k0]['source']
            ][0],
            'lambda0': dout[k0]['lambda0'],
            'pec': f'{k0}_pec',
            'symbol': dout[k0]['symbol'],
            'type': dout[k0]['type'],
            'transition': dout[k0]['transition'],
        }
        for k0 in lk0
    }

    # Get dict of unique ions
    lion = sorted(set([v0['ion'] for v0 in dout.values()]))

    return dne, dte, dpec, lion, dsource, dlines


# #############################################################################
# #############################################################################
#                       Main functions
# #############################################################################


def step03_read(
    adas_path,
    pec_as_func=None,
    **kwdargs,
):
    """ Read openadas-formatted files and return a dict with the data

    Povide the full adas file name
    The result is returned as a dict

    example
    -------
        >>> import spectrally as sp
        >>> fn = '/adf11/scd74/scd74_ar.dat'
        >>> out = sp.openadas.step03_read(fn)
        >>> fn = '/adf15/pec40][ar/pec40][ar_ca][ar16.dat'
        >>> out = sp.openadas.step03_read(fn)
    """

    path_local = _get_PATH_LOCAL()

    # Check whether the local .spectrally repo exists, if not spectrally-custom
    if path_local is None:
        path = os.path.join(os.path.expanduser('~'), f'.{_PKG}', 'openadas')
        msg = (
            f"You do not seem to have a local ./{_PKG} repository\n"
            f"{_PKG} uses that local repository to store all user-specific "
            "data and downloads\n"
            "In particular, openadas files are downloaded and saved in:\n"
            f"\t{path}\n"
            f"  => to set-up your local .{_PKG} repo, run in a terminal:\n"
            f"\t{_PKG}-custom"
        )
        raise Exception(msg)

    # Determine whether adas_path is an absolute path or an adas full name
    if os.path.isfile(adas_path):
        pfe = adas_path
    else:
        # make sure adas_path is not understood as absolute local path
        if adas_path[0] == '/':
            adas_path = adas_path[1:]

        # Check file was downloaded locally
        pfe = os.path.join(path_local, adas_path)
        if not os.path.isfile(pfe):
            msg = (
                "Provided file does not seem to exist:\n"
                f"\t{pfe}\n"
                "  => Search it online with "
                "sp.openadas.step01_search_online()\n"
                "  => Download it with sp.openadas.step02_download()"
            )
            raise FileNotFoundError(msg)

    lc = [ss for ss in _DTYPES.keys() if ss in pfe]
    if not len(lc) == 1:
        msg = (
            "File type could not be derived from absolute path:\n"
            f"\t- provided:  {pfe}\n"
            f"\t- supported: {sorted(_DTYPES.keys())}"
        )
        raise Exception(msg)

    func = eval('_read_{}'.format(lc[0]))
    return func(pfe, **kwdargs)


def step03_read_all(
    element=None, charge=None, typ1=None, typ2=None,
    pec_as_func=None,
    format_for_DataStock=None,
    dsource0=None,
    dref0=None,
    ddata0=None,
    dlines0=None,
    verb=None, **kwdargs,
):
    """ Read all relevant openadas files for chosen typ1

    Please specify:
        - typ1:
            - 'adf11': ionisation / recombination data
            - 'adf15': pec data
        - element: the symbol of the element

    If typ1 = 'adf11', you can also provide typ2 to specify the coefficients:
        - 'scd': effective ionisation coefficients
        - 'acd': effective electron-impact recombination coefficients
        - 'ccd': effective hydrogen-impact recombination coefficients
        - 'plt': line power due to electron-impact excitation
        - 'prc': line power due to hydrogen-impact excitation
        - 'prb': rad. recombination and bremmstrahlung due to electron-impact

    If typ1 = 'adf15', you can optioanlly provide a min/max wavelength

    The result is returned as a dict

    examples
    --------
        >>> import spectrally as sp
        >>> dout = sp.openadas.step03_read_all(element='ar', typ1='adf11')
        >>> dout = sp.openadas.step03_read_all(
            element='ar',
            typ1='adf15',
            charge=16,
            lambmin=3.94e-10,
            lambmax=4.e-10,
        )
    """

    # --------------------
    # Check whether the local .spectrally repo exists, if not spectrally-custom

    path_local = _get_PATH_LOCAL()
    if path_local is None:
        path = os.path.join(os.path.expanduser('~'), f'.{_PKG}', 'openadas')
        msg = (
            f"You do not seem to have a local ./{_PKG} repository\n"
            f"{_PKG} uses that local repository to store all user-specific "
            "data and downloads\n"
            "In particular, openadas files are downloaded and saved in:\n"
            "\t{path}\n"
            f"  => to set-up your local .{_PKG} repo, run in a terminal:\n"
            f"\t{_PKG}-custom"
        )
        raise Exception(msg)

    # --------------------
    # Check / format input
    # --------------------

    if typ1 is None:
        typ1 = 'adf15'
    if not isinstance(typ1, str) or typ1.lower() not in _DTYPES.keys():
        msg = (
            "Please choose a valid adas file type:\n"
            f"\t- allowed:  {_DTYPES.keys()}\n"
            f"\t- provided: {typ1}\n"
        )
        raise Exception(msg)

    typ1 = typ1.lower()
    if typ1 == 'adf11' and typ2 is None:
        typ2 = _DTYPES[typ1]
        fd = os.listdir(os.path.join(path_local, typ1))
        typ2 = [sorted([ss for ss in fd if tt in ss])[-1] for tt in typ2]
    if isinstance(typ2, str):
        typ2 = [typ2]
    if (_DTYPES[typ1] is not None
        and (not isinstance(typ2, list)
             or not all([any([s1 in ss for s1 in _DTYPES[typ1]])
                         for ss in typ2]))):
        msg = ("typ2 must be a list of valid openadas file types for typ1:\n"
               + "\t- provided:         {}\n".format(typ2)
               + "\t- available for {}: {}".format(typ1, _DTYPES[typ1]))
        raise Exception(msg)

    # --------------------
    # Get elevant directory
    # Level 1: Type
    # --------------------

    path = _get_subdir_from_pattern(path_local, typ1, mult='err')[0]

    # --------------------
    # element
    # --------------------

    c0 = (
        element is None
        or isinstance(element, str)
        or (
            isinstance(element, list)
            and all([isinstance(ee, str) for ee in element])
        )
        or (
            isinstance(element, tuple)
            and all([isinstance(ee, str) for ee in element])
        )
    )
    if not c0:
        msg = "Please choose an element!"
        raise Exception(msg)

    if element is None or isinstance(element, tuple):
        el = _get_available_elements_from_path(path=path, typ1=typ1)
        if element is None:
            element = el
        else:
            element = tuple([ee.lower() for ee in element])
            element = [ee for ee in el if ee not in element]
    if isinstance(element, str):
        element = [element]
    element = [ee.lower() for ee in element]

    # --------------------
    # charge
    # --------------------

    if charge is not None:
        c0 = (
            isinstance(charge, int)
            or (
                isinstance(charge, list)
                and all([isinstance(cc, int) for cc in charge])
            )
            or (
                isinstance(charge, tuple)
                and all([isinstance(cc, int) for cc in charge])
            )
        )
        if not c0:
            msg = ("Arg charge must be a int or list (e.g.: 16 or [0])\n"
                   + "\t- provided: {}".format(charge))
            raise Exception(msg)
        if isinstance(charge, int):
            charge = [charge]
        if isinstance(charge, list):
            charge = ['{}.dat'.format(cc) for cc in charge]
        elif isinstance(charge, tuple):
            charge = tuple(['{}.dat'.format(cc) for cc in charge])

    if format_for_DataStock is None:
        format_for_DataStock = False

    if verb is None:
        verb = True

    # --------------------
    # Get list of relevant directories per element
    # --------------------

    # Level 2: element or typ2
    if typ1 == 'adf11':
        lpath = [
            _get_subdir_from_pattern(path, tt, mult='err')[0]
            for tt in typ2
        ]
    elif typ1 == 'adf15':
        lpath = np.concatenate([
            _get_subdir_from_pattern(
                path, ['pec', '][{}'.format(ee)], mult=True,
            )
            for ee in element
        ]).tolist()

    # --------------------
    # Get list of relevant files pfe
    # --------------------

    lpfe = list(itt.chain.from_iterable([[
        os.path.join(path, ff) for ff in os.listdir(path)
        if (
            os.path.isfile(os.path.join(path, ff))
            and ff[-4:] == '.dat'
            and any(['][{}'.format(ee) in ff for ee in element])
        )]
        for path in lpath
    ]))

    if typ1 == 'adf15':
        kwdargs['pec_as_func'] = pec_as_func
        if isinstance(charge, list):
            lpfe = [
                ff for ff in lpfe
                if any([
                    ''.join([ss for ss in ff.split('][')[-1] if ss.isdigit()])
                    + '.dat'
                    == cc
                    for cc in charge
                ])
            ]
        elif isinstance(charge, tuple):
            lpfe = [
                ff for ff in lpfe
                if not any([
                    ''.join([ss for ss in ff.split('][')[-1] if ss.isdigit()])
                    + '.dat'
                    == cc
                    for cc in charge
                ])
            ]

    # --------------------
    # Extract data from each file
    # --------------------

    func = eval(f'_read_{typ1}')
    dout = {}
    for pfe in lpfe:
        if verb is True:
            msg = f"\tLoading data from {pfe}"
            print(msg)
        dout = func(pfe, dout=dout, **kwdargs)

    # format for Collection
    if typ1 == 'adf15' and format_for_DataStock is True:
        return _format_for_DataStock_adf15(
            dout,
            dsource0=dsource0,
            dref0=dref0,
            ddata0=ddata0,
            dlines0=dlines0,
        )

    else:
        return dout


# #############################################################################
#                      Specialized functions for ADF 11
# #############################################################################


def _read_adf11(pfe, deg=None, dout=None):
    if deg is None:
        deg = _DEG
    if dout is None:
        dout = {}

    # Get second order file type
    typ1 = [vv for vv in _DTYPES['adf11'] if vv in pfe]
    if len(typ1) != 1:
        msg = ("Second order file type could not be inferred from file name!\n"
               + "\t- available: {}\n".format(_DTYPES['adf11'])
               + "\t- provided: {}".format(pfe))
        raise Exception(msg)
    typ1 = typ1[0]

    # Get element
    elem = pfe[:-4].split('_')[1]
    comline = '-'*60
    comline2 = 'C'+comline

    if typ1 in ['acd', 'ccd', 'scd', 'plt', 'prb']:

        # read blocks
        with open(pfe) as search:
            for ii, line in enumerate(search):

                if comline2 in line:
                    break

                # Get atomic number (transitions) stored in this file
                if ii == 0:
                    lstr = line.split('/')
                    lin = [ss for ss in lstr[0].strip().split(' ')
                           if ss.strip() != '']
                    lc = [
                        len(lin) == 5 and all([ss.isdigit() for ss in lin]),
                        elem.upper() in lstr[1],
                        # 'ADF11' in lstr[-1],
                    ]
                    if not all(lc):
                        msg = ("File header format seems to have changed!\n"
                               + "\t- pfe: {}\n".format(pfe)
                               + "\t- lc = {}\n".format(lc)
                               + "\t- lstr = {}".format(lstr))
                        raise Exception(msg)
                    Z, nne, nte, q0, qend = map(int, lin)
                    nelog10 = np.array([])
                    telog10 = np.array([])
                    in_ne = True
                    continue

                if comline in line:
                    continue

                # Get nelog10
                if in_ne:
                    li = [ss for ss in line.strip().split(' ')
                          if ss.strip() != '']
                    nelog10 = np.append(nelog10, np.array(li, dtype=float))
                    if nelog10.size == nne:
                        in_ne = False
                        in_te = True

                # Get telog10
                elif in_te is True:
                    li = [ss for ss in line.strip().split(' ')
                          if ss.strip() != '']
                    telog10 = np.append(telog10, np.array(li, dtype=float))
                    if telog10.size == nte:
                        in_te = False
                        in_ion = True

                # Get ion block
                elif (in_ion is True and 'Z1=' in line
                      and ('------/' in line and 'DATE=' in line)):
                    nion = int(
                        line[line.index('Z1=')+len('Z1='):].split('/')[0])
                    if typ1 in ['scd', 'plt']:
                        charge = nion - 1
                    else:
                        charge = nion
                    coefslog10 = np.array([])
                elif in_ion is True and charge is not None:
                    li = [ss for ss in line.strip().split(' ')
                          if ss.strip() != '']
                    coefslog10 = np.append(coefslog10,
                                           np.array(li, dtype=float))
                    if coefslog10.size == nne*nte:
                        key = '{}{}'.format(elem, charge)
                        tkv = [('element', elem), ('Z', Z), ('charge', charge)]
                        if key in dout.keys():
                            assert all([dout[key][ss] == vv for ss, vv in tkv])
                        else:
                            dout[key] = {ss: vv for ss, vv in tkv}
                        if typ1 == 'scd':
                            # nelog10+6 to convert /cm3 -> /m3
                            # coefslog10-6 to convert cm3/s -> m3/s
                            func = scpRectSpl(nelog10+6, telog10,
                                              coefslog10.reshape((nne, nte))-6,
                                              kx=deg, ky=deg)
                            dout[key]['ionis'] = {'func': func,
                                                  'type': 'log10_nete',
                                                  'units': 'log10(m3/s)',
                                                  'source': pfe}
                        elif typ1 == 'acd':
                            # nelog10+6 to convert /cm3 -> /m3
                            # coefslog10-6 to convert cm3/s -> m3/s
                            func = scpRectSpl(nelog10+6, telog10,
                                              coefslog10.reshape((nne, nte))-6,
                                              kx=deg, ky=deg)
                            dout[key]['recomb'] = {'func': func,
                                                   'type': 'log10_nete',
                                                   'units': 'log10(m3/s)',
                                                   'source': pfe}
                        elif typ1 == 'ccd':
                            # nelog10+6 to convert /cm3 -> /m3
                            # coefslog10-6 to convert cm3/s -> m3/s
                            func = scpRectSpl(nelog10+6, telog10,
                                              coefslog10.reshape((nne, nte))-6,
                                              kx=deg, ky=deg)
                            dout[key]['recomb_ce'] = {'func': func,
                                                      'type': 'log10_nete',
                                                      'units': 'log10(m3/s)',
                                                      'source': pfe}
                        elif typ1 == 'plt':
                            # nelog10+6 to convert /cm3 -> /m3
                            # coefslog10+6 to convert W.cm3 -> W.m3
                            func = scpRectSpl(nelog10+6, telog10,
                                              coefslog10.reshape((nne, nte))+6,
                                              kx=deg, ky=deg)
                            dout[key]['rad_bb'] = {'func': func,
                                                   'type': 'log10_nete',
                                                   'units': 'log10(W.m3)',
                                                   'source': pfe}
                        elif typ1 == 'prb':
                            # nelog10+6 to convert /cm3 -> /m3
                            # coefslog10+6 to convert W.cm3 -> W.m3
                            func = scpRectSpl(nelog10+6, telog10,
                                              coefslog10.reshape((nne, nte))+6,
                                              kx=deg, ky=deg)
                            dout[key]['rad_fffb'] = {'func': func,
                                                     'type': 'log10_nete',
                                                     'units': 'log10(W.m3)',
                                                     'source': pfe}
                        if nion == Z:
                            break

    return dout


# #############################################################################
#                      Specialized functions for ADF 15
# #############################################################################


def _get_adf15_key(elem, charge, isoel, typ0, typ1):
    return '{}{}_{}_oa_{}_{}'.format(elem, charge, isoel, typ0, typ1)


def _read_adf15(
    pfe,
    dout=None,
    lambmin=None,
    lambmax=None,
    pec_as_func=None,
    deg=None,
):
    """
    Here lambmin and lambmax are provided in m
    """

    if deg is None:
        deg = _DEG
    if pec_as_func is None:
        pec_as_func = _PECASFUNC
    if dout is None:
        dout = {}

    # Get summary of transitions
    flagblock = '/isel ='
    flag0 = 'superstage partition information'

    # Get file markers from name (elem, charge, typ0, typ1)
    typ0, typ1, elemq = pfe.split('][')[1:]
    ind = re.search(r'\d', elemq).start()
    elem = elemq[:ind].title()
    charge = int(elemq[ind:-4])
    assert elem.lower() in typ0[:2]
    assert elem.lower() == typ1.split('_')[0]
    typ0 = typ0[len(elem)+1:]
    typ1 = typ1.split('_')[1]

    # Extract data from file
    nlines, nblock = None, 0
    in_ne, in_te, in_pec, in_tab, itab = False, False, False, False, np.inf
    skip = False
    with open(pfe) as search:
        for ii, line in enumerate(search):

            # Get number of lines (transitions) stored in this file
            if ii == 0:
                lstr = line.split('/')
                nlines = int(lstr[0].replace(' ', ''))
                continue

            # Get info about the transition being scanned (block)
            if flagblock in line and 'C' not in line and nblock < nlines:
                lstr = [kk for kk in line.rstrip().split(' ') if len(kk) > 0]
                lamb = float(lstr[0])*1.e-10
                isoel = nblock + 1
                nblock += 1
                c0 = ((lambmin is not None and lamb < lambmin)
                      or (lambmax is not None and lamb > lambmax))
                if c0:
                    skip = True
                    continue
                skip = False
                nne, nte = int(lstr[1]), int(lstr[2])
                typ = [ss[ss.index('type=')+len('type='):ss.index('/ispb')]
                       for ss in lstr[3:] if 'type=' in ss]
                assert len(typ) == 1
                # To be updated : proper reading from line
                in_ne = True
                ne = np.array([])
                te = np.array([])
                pec = np.full((nne*nte,), np.nan)
                ind = 0
                # il = 0
                continue

            if 'root partition information' in line and skip is True:
                skip = False

            # Check lamb is ok
            if skip is True:
                continue

            # Get ne for the transition being scanned (block)
            if in_ne is True:
                ne = np.append(ne,
                               np.array(line.rstrip().strip().split(' '),
                                        dtype=float))
                if ne.size == nne:
                    in_ne = False
                    in_te = True

            # Get te for the transition being scanned (block)
            elif in_te is True:
                te = np.append(te,
                               np.array(line.rstrip().strip().split(' '),
                                        dtype=float))
                if te.size == nte:
                    in_te = False
                    in_pec = True

            # Get pec for the transition being scanned (block)
            elif in_pec is True:
                data = np.array(line.rstrip().strip().split(' '),
                                dtype=float)
                pec[ind:ind+data.size] = data
                # pec[il, :] = data
                ind += data.size
                # il += 1
                if ind == pec.size:
                    in_pec = False
                    key = _get_adf15_key(elem, charge, isoel, typ0, typ1)
                    # PEC reshaping and conversion to cm3/s -> m3/s
                    pec = pec.reshape((nne, nte)) * 1e-6
                    # log(ne)+6 to convert /cm3 -> /m3
                    ne = ne*1e6

                    if pec_as_func is True:
                        pec_rec = scpRectSpl(
                            np.log(ne),
                            np.log(te),
                            np.log(pec),
                            kx=deg,
                            ky=deg,
                        )

                        def pec(Te=None, ne=None, pec_rec=pec_rec):
                            return np.exp(pec_rec(np.log(ne), np.log(Te)))

                    dout[key] = {
                        'lambda0': lamb,
                        'ion': '{}{}+'.format(elem, charge),
                        'charge': charge,
                        'element': elem,
                        'symbol': '{}{}-{}'.format(typ0, typ1, isoel),
                        'source': pfe,
                        'type': typ[0],
                        'ne': ne,
                        'ne_units': '/m3',
                        'te': te,
                        'te_units': 'eV',
                        'pec': pec,
                        'pec_type': 'f(ne, Te)',
                        'pec_units': 'm3/s',
                    }

            # Get transitions from table at the end
            if 'photon emissivity atomic transitions' in line:
                itab = ii + 6
            if ii == itab:
                in_tab = True
            if in_tab is True:
                lstr = [kk for kk in line.rstrip().split(' ') if len(kk) > 0]
                isoel = int(lstr[1])
                lamb = float(lstr[2])*1.e-10
                key = _get_adf15_key(elem, charge, isoel, typ0, typ1)
                c0 = ((lambmin is None or lambmin < lamb)
                      and (lambmax is None or lambmax > lamb))
                if c0 and key not in dout.keys():
                    msg = ("Inconsistency in file {}:\n".format(pfe)
                           + "\t- line should be present".format(key))
                    raise Exception(msg)
                if key in dout.keys():
                    if dout[key]['lambda0'] != lamb:
                        msg = "Inconsistency in file {}".format(pfe)
                        raise Exception(msg)
                    c0 = (dout[key]['type'] not in lstr
                          or lstr.index(dout[key]['type']) < 4)
                    if c0:
                        msg = ("Inconsistency in table, type not found:\n"
                               + "\t- expected: {}\n".format(dout[key]['type'])
                               + "\t- line: {}".format(line))
                        raise Exception(msg)
                    trans = lstr[3:lstr.index(dout[key]['type'])]
                    dout[key]['transition'] = ''.join(trans)
                if isoel == nlines:
                    in_tab = False
    assert all(['transition' in vv.keys() for vv in dout.values()])
    return dout



# -*- coding: utf-8 -*-
"""
Created on Sat Mar  9 16:09:08 2024

@author: dvezinet
"""


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
import matplotlib.colors as mcolors
import matplotlib.transforms as transforms
import datastock as ds


# local


#############################################
#############################################
#       DEFAULTS
#############################################


_MK = '.'
_MS = 6
_DPROP = {
    '0':  {'ls': 'None', 'marker': _MK, 'ms': _MS},
    '-1': {'ls': 'None', 'marker': _MK, 'ms': _MS},
    '-2': {'ls': 'None', 'marker': _MK, 'ms': _MS},
    '-3': {'ls': 'None', 'marker': _MK, 'ms': _MS},
    '-4': {'ls': 'None', 'marker': _MK, 'ms': _MS},
    '-5': {'ls': 'None', 'marker': _MK, 'ms': _MS},
    '-6': {'ls': 'None', 'marker': _MK, 'ms': _MS},
    '-7': {'ls': 'None', 'marker': _MK, 'ms': _MS},
    '-8': {'ls': 'None', 'marker': _MK, 'ms': _MS},
}


def set_dprop():
    lk = sorted(_DPROP.keys())
    lk_int = np.array(lk, dtype=int)
    vmin = np.min(lk_int)
    vmax = np.max(lk_int)
    norm = mcolors.Normalize(vmin=vmin, vmax=vmax)
    cm = plt.cm.Set1
    for k0 in _DPROP.keys():
        _DPROP[k0]['color'] = cm(norm(int(k0)))

    return cm, vmin, vmax

_CM, _VMIN, _VMAX = set_dprop()


#############################################
#############################################
#       main
#############################################


def plot(
    coll=None,
    key=None,
    keyY=None,
    dref_vectorY=None,
    # options
    dprop=None,
    vmin=None,
    vmax=None,
    cmap=None,
    plot_text=None,
    # figure
    dax=None,
    fs=None,
    dmargin=None,
    tit=None,
    # interactivity
    connect=True,
    dinc=None,
    show_commands=None,
):

    # -----------------
    # check
    # -----------------

    key, keyY, keyZ, refY, refZ, dprop, vmin, vmax, tit = _check(
        coll=coll,
        key=key,
        # keyY
        keyY=keyY,
        dref_vectorY=dref_vectorY,
        # options
        dprop=dprop,
        vmin=vmin,
        vmax=vmax,
        plot_text=plot_text,
        # figure
        tit=tit,
    )

    wsf = coll._which_fit
    data = coll.ddata[coll.dobj[wsf][key]['key_data']]['data']
    ndim = data.ndim
    key_bs = coll.dobj[wsf][key]['key_bs']
    key_bs_vect = coll.dobj[wsf][key]['key_bs_vect']

    # -----------------
    # plot
    # -----------------

    if ndim == 1:
        assert key_bs is None

        dax = _plot_1d(
            coll=coll,
            key=key,
            # options
            dprop=dprop,
            vmin=vmin,
            vmax=vmax,
            # figure
            dax=dax,
            fs=fs,
            dmargin=dmargin,
        )

    elif ndim == 2:

        dax, dgroup = _plot_2d(
            coll=coll,
            key=key,
            # keyY
            keyY=keyY,
            refY=refY,
            # bsplines
            key_bs=key_bs,
            key_bs_vect=key_bs_vect,
            # options
            dprop=dprop,
            dvminmax=None,
            cmap=cmap,
            plot_text=plot_text,
            # figure
            dax=dax,
            fs=fs,
            dmargin=dmargin,
        )

    else:
        raise NotImplementedError()

    # -----------------
    # Complenent
    # -----------------

    if isinstance(dax, dict):
        fig = list(dax.values())[0]['handle'].figure
    else:
        fig = list(dax.dax.values())[0]['handle'].figure

    if tit is not False:
        fig.suptitle(tit, size=14, fontweight='bold')

    # ----------------------
    # connect interactivity
    # ----------------------

    if ndim == 1:
        return dax

    elif connect is True:
        dax.setup_interactivity(
            kinter='inter0', dgroup=dgroup, dinc=dinc,
        )
        dax.disconnect_old()
        dax.connect()

        dax.show_commands(verb=show_commands)
        return dax

    else:
        return dax, dgroup


#############################################
#############################################
#       check
#############################################


def _check(
    coll=None,
    key=None,
    # keyY
    keyY=None,
    dref_vectorY=None,
    # keyZ
    keyZ=None,
    dref_vectorZ=None,
    # options
    dprop=None,
    vmin=None,
    vmax=None,
    plot_text=None,
    # figure
    tit=None,
):

    # -----------------
    # key
    # -----------------

    wsf = coll._which_fit
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=list(coll.dobj.get(wsf, {}).keys()),
    )
    key_data = coll.dobj[wsf][key]['key_data']
    key_lamb = coll.dobj[wsf][key]['key_lamb']
    ref_lamb = coll.ddata[key_lamb]['ref'][0]

    refs = coll.ddata[key_data]['ref']

    # -----------------
    # keyY
    # -----------------

    ndim = coll.ddata[coll.dobj[wsf][key]['key_data']]['data'].ndim
    key_bs_vect = coll.dobj[wsf][key]['key_bs_vect']

    keyZ = None
    refZ = None
    if ndim == 1:
        keyY = None
        refY = None

    elif ndim == 2:
        refY = refs[1-refs.index(ref_lamb)]

        if key_bs_vect is None:

            if dref_vectorY is None:
                dref_vectorY = {}
            if dref_vectorY.get('key0') is None:
                dref_vectorY['key0'] = keyY

            if ndim == 2 and dref_vectorY.get('ref') is None:
                ref = [
                    rr for rr in coll.ddata[key_data]['ref']
                    if rr != ref_lamb
                ][0]
                dref_vectorY['ref'] = ref

            keyY = coll.get_ref_vector(**dref_vectorY)[3]

            Ydatadiff = np.diff(coll.ddata[keyY]['data'])
            if not np.allclose(Ydatadiff, Ydatadiff[0], rtol=1e-6, atol=0):
                keyY = None

        else:
            keyY = key_bs_vect

    else:
        raise NotImplementedError()

    # -----------------
    # dprop
    # -----------------

    if dprop is None:
        dprop = {}

    if not isinstance(dprop, dict):
        msg = "Arg dprop must be a dict"
        raise Exception(msg)

    lk = coll.dobj[wsf][key]['dvalid']['meaning'].keys()
    for k0 in lk:

        if dprop.get(k0) is None:
            dprop[k0] = {}

        for k1, v1 in _DPROP[k0].items():
            if dprop[k0].get(k1) is None:
                dprop[k0][k1] = _DPROP[k0][k1]

    # -----------------
    # vmin, vmax
    # -----------------

    # vmin
    vmin = float(ds._generic_check._check_var(
        vmin, 'vmin',
        types=(int, float),
        default=0,
    ))

    # vmax
    key_data = coll.dobj[wsf][key]['key_data']
    vmax_def = np.nanmax(coll.ddata[key_data]['data']) * 1.05
    vmax = float(ds._generic_check._check_var(
        vmax, 'vmax',
        types=(int, float),
        default=vmax_def,
    ))

    # -----------------
    # plot_text
    # -----------------

    plot_text = ds._generic_check._check_var(
        plot_text, 'plot_text',
        types=bool,
        default=False,
    )

    # -----------------
    # figure
    # -----------------

    tit_def = f"input data validity for {wsf} '{key}'"
    tit = ds._generic_check._check_var(
        tit_def, 'tit_def',
        types=(str, bool),
        default=tit_def,
    )

    return key, keyY, keyZ, refY, refZ, dprop, vmin, vmax, tit


#############################################
#############################################
#       plot 1d
#############################################


def _plot_1d(
    coll=None,
    key=None,
    # options
    dprop=None,
    vmin=None,
    vmax=None,
    # figure
    dax=None,
    fs=None,
    dmargin=None,
):

    # -----------------
    # prepare
    # -----------------

    wsf = coll._which_fit
    lamb = coll.ddata[coll.dobj[wsf][key]['key_lamb']]['data']
    data = coll.ddata[coll.dobj[wsf][key]['key_data']]['data']

    dvalid = coll.dobj[wsf][key]['dvalid']
    iok = coll.ddata[dvalid['iok']]['data']
    frac = dvalid['frac'][0]

    # -----------------
    # prepare figure
    # -----------------

    if dax is None:
        dax = _get_dax_1d(
            fs=fs,
            dmargin=dmargin,
            # labelling
            coll=coll,
            key=key,
        )

    dax = ds._generic_check._check_dax(dax=dax, main='data')

    # -----------------
    # plot
    # -----------------

    kax = 'data'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # validity
        for k0, v0 in dvalid['meaning'].items():

            ind = (iok == int(k0))
            ax.plot(
                lamb[ind],
                data[ind],
                label=v0,
                **dprop[k0],
            )

        # legend
        ax.legend(loc='upper left', bbox_to_anchor=(1, 1), fontsize=10)

        # frac
        ax.set_title(
            f"frac = {frac:.3f} vs {dvalid['fraction']}",
            size=12,
            fontweight='bold',
        )

        # nsigma
        ax.axhline(dvalid['nsigma']**2, ls='--', c='k')

        trans = transforms.blended_transform_factory(
            ax.transAxes,
            ax.transData,
        )
        ax.text(
            1.02,
            dvalid['nsigma']**2,
            r'$n_{\sigma}^2$' + f" = {dvalid['nsigma']}" + r"$^2$",
            size=12,
            fontweight='normal',
            transform=trans,
        )

        # focus
        if dvalid.get('focus') is not None:
            for ff in dvalid['focus']:
                ax.axvspan(ff[0], ff[1], fc=(0.8, 0.8, 0.8, 0.5))

        # vmin vmax
        if vmin is not None:
            ax.set_ylim(bottom=vmin)
        if vmax is not None:
            ax.set_ylim(top=vmax)

    return dax


# ---------------------
# create axes
# ---------------------


def _get_dax_1d(
    fs=None,
    dmargin=None,
    # labelling
    coll=None,
    key=None,
):

    # ---------------
    # check
    # ---------------

    if fs is None:
        fs = (11, 6)

    if dmargin is None:
        dmargin = {
            'left': 0.07, 'right': 0.78,
            'bottom': 0.08, 'top': 0.90,
            'wspace': 0.10, 'hspace': 0.10,
        }

    # ---------------
    # prepare labels
    # ---------------

    wsf = coll._which_fit
    key_lamb = coll.dobj[wsf][key]['key_lamb']
    key_data = coll.dobj[wsf][key]['key_data']
    xlab = f"{key_lamb} ({coll.ddata[key_lamb]['units']})"
    ylab = f"{key_data} ({coll.ddata[key_data]['units']})"

    # ---------------
    # figure
    # ---------------

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(ncols=1, nrows=1, **dmargin)

    # ---------------
    # axes
    # ---------------

    ax = fig.add_subplot(gs[0, 0])
    ax.set_xlabel(xlab, size=12, fontweight='bold')
    ax.set_ylabel(ylab, size=12, fontweight='bold')

    dax = {'data': {'handle': ax}}

    return dax


#############################################
#############################################
#       plot 2d
#############################################


def _plot_2d(
    coll=None,
    key=None,
    # keyY
    keyY=None,
    refY=None,
    # bsplines
    key_bs=None,
    key_bs_vect=None,
    # options
    dprop=None,
    dvminmax=None,
    cmap=None,
    plot_text=None,
    # figure
    dax=None,
    fs=None,
    dmargin=None,
):

    # -----------------
    # prepare
    # -----------------

    wsf = coll._which_fit
    key_data = coll.dobj[wsf][key]['key_data']
    key_lamb = coll.dobj[wsf][key]['key_lamb']

    dvalid = coll.dobj[wsf][key]['dvalid']

    # -----------------
    # prepare figure
    # -----------------

    if dax is None:
        dax = _get_dax_2d(
            fs=fs,
            dmargin=dmargin,
            # bsplines
            keyY=keyY,
            key_bs=key_bs,
            key_bs_vect=key_bs_vect,
            dmeaning=dvalid['meaning'],
            # labelling
            coll=coll,
            key=key,
        )

    dax = ds._generic_check._check_dax(dax=dax, main='data')

    # -----------------
    # prepare - extract relevant data
    # -----------------

    collax, key_data, key_iok, key_lamb, key_frac = _prepare(
        coll=coll,
        key=key,
        keyY=keyY,
    )

    # -----------------
    # plot as array
    # -----------------

    lax = ['data_img', 'data_vert', 'spectrum']
    if plot_text is True:
        lax += ['text_Y']

    collax, dgroup = collax.plot_as_array(
        key=key_data,
        keyX=key_lamb,
        keyY=keyY,
        aspect='auto',
        dvminmax=dvminmax,
        cmap=cmap,
        dax={k0: v0 for k0, v0 in dax.items() if k0 in lax},
        inplace=True,
        connect=False,
    )

    # ---------------------
    # adjust keyY if needed
    # ---------------------

    if keyY is None:
        kmob = f"{key_data}_h00"
        keyY = collax.dobj['mobile'][kmob]['data'][0]
        refY = collax.ddata[keyY]['ref'][0]

    # -----------------
    # plot valid
    # -----------------

    collax, dgroup2 = collax.plot_as_array(
        key=key_iok,
        keyX=key_lamb,
        keyY=keyY,
        aspect='auto',
        dvminmax={'data': {'min': _VMIN, 'max': _VMAX}},
        cmap=_CM,
        dax={
            k0: v0 for k0, v0 in dax.items()
            if k0 in ['valid_img', 'valid_1d']
        },
        inplace=True,
        connect=False,
    )

    # ----------------
    # add nsigma and focus
    # ---------------

    kax = 'spectrum'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # hline
        ax.axhline(dvalid['nsigma']**2, ls='--', c='k')

        # text
        trans = transforms.blended_transform_factory(
            ax.transAxes,
            ax.transData,
        )
        ax.text(
            1.02,
            dvalid['nsigma']**2,
            r'$n_{\sigma}^2$' + f" = {dvalid['nsigma']}" + r"$^2$",
            size=12,
            fontweight='normal',
            transform=trans,
        )

        # focus
        if dvalid.get('focus') is not None:
            for ff in dvalid['focus']:
                ax.axvspan(ff[0], ff[1], fc=(0.8, 0.8, 0.8, 0.5))

    # -----------------
    # plot frac
    # -----------------

    kax = 'frac'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        dcol = {
            int(k0.split('_')[-1][1:]): v0['handle'].get_color()
            for k0, v0 in collax.dobj['mobile'].items()
            if k0.startswith(f'{key_iok}_h')
        }

        # ind0, ind1
        for ii in range(dgroup['Y']['nmax']):

            lh = ax.axhline(
                0,
                c=dcol[ii],
                lw=1.,
                ls='-',
            )

            # update collax
            kh = f'frac_h{ii}'
            collax.add_mobile(
                key=kh,
                handle=lh,
                refs=refY, # collax.ddata[keyY]['ref'],
                data=(keyY,),
                dtype='ydata',
                axes=kax,
                ind=ii,
            )

        # fixed
        ax.axvline(
            dvalid['fraction'],
            c='k',
            ls='--',
            lw=1,
        )

        # fixed
        ax.plot(
            collax.ddata[dvalid['frac']]['data'],
            collax.ddata[keyY]['data'],
            c='k',
            ls='-',
            lw=1,
        )

        # axes interactivity
        # add axes
        collax.add_axes(
            handle=ax,
            key=kax,
            refy=[refY],
            datay=[keyY],
            harmonize=True,
        )

    return collax, dgroup


# ---------------------
# prepare data extraction
# ---------------------


def _prepare(
    coll=None,
    key=None,
    keyY=None,
    keyZ=None,
):

    # ----------
    # prepare
    # ----------

    # data, lamb
    wsf = coll._which_fit
    key_data = coll.dobj[wsf][key]['key_data']
    key_lamb = coll.dobj[wsf][key]['key_lamb']
    ndim = coll.ddata[key_data]['data'].ndim

    # iok, frac
    dvalid = coll.dobj[wsf][key]['dvalid']
    key_iok = dvalid['iok']
    key_frac = dvalid['frac']

    # keyY

    # ----------
    # list keys
    # ----------

    keys = [key_data, key_lamb, key_iok, key_frac]

    if keyY is not None:
        keys.append(keyY)
    if ndim == 3:
        keys.append(keyZ)

    # ---------
    # extract
    # ---------

    collax = coll.extract(
        keys=keys,
        inc_monot=True,
        inc_vectors=True,
        inc_allrefs=False,
        return_keys=False,
    )[0]

    return collax, key_data, key_iok, key_lamb, key_frac


# ---------------------
# create axes
# ---------------------


def _get_dax_2d(
    fs=None,
    dmargin=None,
    keyY=None,
    # bsplines
    key_bs=None,
    key_bs_vect=None,
    dmeaning=None,
    # labelling
    coll=None,
    key=None,
):

    # ---------------
    # check
    # ---------------

    if fs is None:
        fs = (14, 10)

    if dmargin is None:
        dmargin = {
            'left': 0.06, 'right': 0.98,
            'bottom': 0.06, 'top': 0.92,
            'wspace': 0.9, 'hspace': 0.50,
        }

    # ---------------
    # prepare labels
    # ---------------

    wsf = coll._which_fit
    key_lamb = coll.dobj[wsf][key]['key_lamb']
    key_data = coll.dobj[wsf][key]['key_data']
    xlab = f"{key_lamb} ({coll.ddata[key_lamb]['units']})"
    dlab = f"{key_data} ({coll.ddata[key_data]['units']})"
    if keyY is None:
        yunits = ''
    else:
        yunits = coll.ddata[keyY]['units']
    ylab = f"{keyY} ({yunits})"

    # ---------------
    # figure
    # ---------------

    fig = plt.figure(figsize=fs)

    nh = (1, 3, 9)
    ncols = (2 * nh[0] + 2 * nh[1] + 2 * nh[2])
    gs = gridspec.GridSpec(ncols=ncols, nrows=6, **dmargin)

    # ---------------
    # axes
    # ---------------

    dax = {}

    # data_img
    n0 = nh[0]
    ax = fig.add_subplot(gs[:3, n0:(n0 + nh[2])])
    ax.set_ylabel(ylab, size=12, fontweight='bold')
    ax.set_title('data', size=14, fontweight='bold')
    dax['data_img'] = {'handle': ax, 'type': 'matrix'}
    ax0 = ax

    # data_cbar
    ax = fig.add_subplot(gs[:3, :nh[0]])
    dax['data_cbar'] = ax

    # data_vert
    n0 = nh[0] + nh[2]
    ax = fig.add_subplot(gs[:3, n0:(n0+nh[1])], sharey=ax0)
    ax.set_xlabel(dlab, size=12, fontweight='bold')
    ax.set_ylabel(ylab, size=12, fontweight='bold')
    dax['data_vert'] = {'handle': ax, 'type': 'vertical'}

    # bs_knots
    if key_bs_vect is not None:
        n0 = (nh[0] + nh[1] + nh[2])
        ax = fig.add_subplot(gs[:3, n0:(n0+nh[0])], sharey=ax0)
        dax['bs_knots'] = ax

    # valid_img
    n0 = (2*nh[0] + 2*nh[1] + nh[2])
    ax = fig.add_subplot(gs[:3, n0:(n0+nh[2])], sharex=ax0, sharey=ax0)
    ax.set_xlabel(xlab, size=12, fontweight='bold')
    ax.set_ylabel(ylab, size=12, fontweight='bold')
    ax.set_title('validity', size=14, fontweight='bold')
    dax['valid_img'] = {'handle': ax, 'type': 'matrix'}

    # frac
    n0 = (2*nh[0] + nh[1] + nh[2])
    ax = fig.add_subplot(gs[:3, n0:(n0+nh[1])], sharey=ax0)
    ax.set_title('frac', size=14, fontweight='bold')
    dax['frac'] = {'handle': ax, 'type': 'frac'}

    # spectrum
    n0 = nh[0]
    ax = fig.add_subplot(gs[3:5, n0:(n0+nh[2])], sharex=ax0)
    ax.set_ylabel(dlab, size=12, fontweight='bold')
    ax.set_title('spectrum', size=14, fontweight='bold')
    dax['spectrum'] = {'handle': ax, 'type': 'horizontal'}

    # valid_1d
    n0 = nh[0]
    ax = fig.add_subplot(gs[5, n0:(n0+nh[2])], sharex=ax0)
    ax.set_xlabel(xlab, size=12, fontweight='bold')
    ax.set_ylabel('validity', size=12, fontweight='bold')
    trans = transforms.blended_transform_factory(
        ax.transAxes,
        ax.transData,
    )
    for k0, v0 in dmeaning.items():
        ax.text(
            1.02,
            int(k0),
            v0,
            color=_DPROP[k0]['color'],
            transform=trans,
            horizontalalignment='left',
            verticalalignment='center',
        )
    dax['valid_1d'] = {'handle': ax, 'type': 'horizontal'}

    # text - refY
    n0 = (2*nh[0] + nh[1] + nh[2])
    ax = fig.add_subplot(gs[3:, n0:(n0+2*nh[1])])
    ax.axis('off')
    dax['text_Y'] = {'handle': ax, 'type': 'textY'}

    # -----------------
    # adjust visibility
    # -----------------

    plt.setp(dax['data_img']['handle'].get_xticklabels(), visible=False)
    plt.setp(dax['spectrum']['handle'].get_xticklabels(), visible=False)
    plt.setp(dax['spectrum']['handle'].get_xlabel(), visible=False)

    return dax


#############################################
#############################################
#       plot 2d
#############################################